# Copyright 2023-2024 The SapientML Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import re
from collections import defaultdict
from typing import Annotated, Any, Literal, Optional, Union

import numpy as np
import pandas as pd
from pandas.core.dtypes.common import is_numeric_dtype
from pydantic import BaseModel, Field, StringConstraints, field_validator
from sapientml.params import (
    INITIAL_TIMEOUT,
    MAX_COLUMN_NAME_LENGTH,
    MAX_NUM_OF_COLUMNS,
    Code,
    ColumnName,
    Config,
    Seed,
    String,
    Task,
)

from . import ps_macros
from .meta_features import (
    MetaFeatures,
    generate_column_meta_features,
    generate_model_meta_features,
    generate_pp_meta_features,
)

PipelineSkeleton = (
    dict  # dict[str, Union[float, dict[str, Union[float, list[str], list[dict[str, Union[float, int, str]]]]]]]
)

MAX_N_MODELS = 30
MAX_HPO_N_TRIALS = 100000
MAX_HPO_TIMEOUT = 500000

NModels = Annotated[int, Field(strict=False, gt=0, le=MAX_N_MODELS)]
TuningNTrials = Annotated[int, Field(strict=False, ge=1, le=MAX_HPO_N_TRIALS)]
TuningTimeout = Annotated[int, Field(strict=False, ge=0, le=MAX_HPO_TIMEOUT)]
Dtype = Annotated[str, StringConstraints(max_length=100)]


class SapientMLConfig(Config):
    """
    Configuration arguments for SapientMLGenerator.

    Attributes
    ----------
    n_models: int, default 3
        Number of output models to be tried.
    seed_for_model: int, default 42
        Random seed for models such as RandomForestClassifier.
    id_columns_for_prediction: Optional[list[str]], default None
        Name of the dataframe columns that outputs the prediction result.
    use_word_list: Optional[Union[list[str], dict[str, list[str]]]], default None
        List of words to be used as features when generating explanatory variables from text.
        If dict type is specified, key must be a column name and value must be a list of words.
    hyperparameter_tuning: bool, default False
        On/Off of hyperparameter tuning.
    hyperparameter_tuning_n_trials: int, default 10
        The number of trials of hyperparameter tuning.
    hyperparameter_tuning_timeout: int, default 0
        Time limit for hyperparameter tuning in each generated script.
        Ignored when hyperparameter_tuning is False.
    hyperparameter_tuning_random_state: int, default 1023
        Random seed for hyperparameter tuning.
    predict_option: Literal["default", "probability", None], default None
        Specify predict method (default: predict(), probability: predict_proba(), None: Comply with metric requirements.)
    permutation_importance: bool, default True
        On/Off of outputting permutation importance calculation code.
    add_explanation: bool, default False
        If True, outputs ipynb files including EDA and explanation.
    export_preprocess_dataset : bool, default False
        export pre-processed (feature engineered) the input dataset


    """

    n_models: NModels = 3
    seed_for_model: Seed = 42
    id_columns_for_prediction: Optional[list[ColumnName]] = None
    use_word_list: Optional[Union[list[ColumnName], dict[str, list[ColumnName]]]] = None
    hyperparameter_tuning: bool = False
    hyperparameter_tuning_n_trials: TuningNTrials = 10
    hyperparameter_tuning_timeout: TuningTimeout = 0
    hyperparameter_tuning_random_state: Seed = 1023
    predict_option: Optional[Literal["default", "probability"]] = None
    permutation_importance: bool = True
    add_explanation: bool = False
    export_preprocess_dataset: bool = False

    def post_init(self):
        """Set initial_timeout and hyperparameter_tuning_timeout.

        If initial_timeout is set as None and hyperparameter_tuning is false, set initial_timeout as INITIAL_TIMEOUT.

        For hyperparameter_tuning_timeout,
        if both initial_timeout and hyperparameter_tuning_timeout are set as None, set hyperparameter_tuning_timeout as INITIAL_TIMEOUT.

        If initial_timeout is set and hyperparameter_tuning is True,
        and hyperparameter_tuning_timeout is None :
            Set the hyperparameter_tuning_timeout to unlimited.(hyperparameter_tuning_timeout = self.initial_timeout.)
            Since initial_timeout always precedes hyperparameter_tuning_timeout,
            it can be expressed that there is no time limit for hyperparameters during actual execution.
        """
        if self.id_columns_for_prediction is None:
            self.id_columns_for_prediction = []

        if self.initial_timeout is None:
            if self.hyperparameter_tuning:
                self.initial_timeout = 0
            else:
                self.initial_timeout = INITIAL_TIMEOUT
            if self.hyperparameter_tuning_timeout is None:
                self.hyperparameter_tuning_timeout = INITIAL_TIMEOUT
        elif self.hyperparameter_tuning_timeout is None:
            if self.hyperparameter_tuning:
                self.hyperparameter_tuning_timeout = self.initial_timeout
            else:
                self.hyperparameter_tuning_timeout = INITIAL_TIMEOUT

    @field_validator(
        "id_columns_for_prediction",
        "use_word_list",
    )
    def _check_num_of_column_names(cls, v):
        if v is None:
            return v
        if len(v.keys() if isinstance(v, dict) else v) >= MAX_NUM_OF_COLUMNS:
            raise ValueError(f"The number of columns must be smaller than {MAX_NUM_OF_COLUMNS}")
        return v


class Column(BaseModel):
    """
    Describing meta-features of a column in the input dataset.

    Attributes
    ----------
    dtype : str
        Data type of the column of features.
    meta_features : MetaFeatures | None
        Meta features of the column.
    has_negative_value : bool
        Whether the column has a negative value.
    """

    dtype: Dtype
    meta_features: Optional[MetaFeatures]
    has_negative_value: bool

    @field_validator("meta_features")
    def check_meta_features(cls, v):
        if v is None:
            return v
        for k, _v in v.items():
            if not re.match(r"feature:[a-z_0-9]+", k):
                raise ValueError(f"'{k}' is invalid as a feature name")
            if isinstance(_v, int):
                pass
            elif isinstance(_v, float):
                pass
            elif isinstance(_v, str) and len(_v) >= MAX_COLUMN_NAME_LENGTH:
                raise ValueError(f"Meta feature string value must be shorter than {MAX_COLUMN_NAME_LENGTH}")
            elif isinstance(_v, list):
                for s in _v:
                    if not isinstance(s, str):
                        raise ValueError("The list must contain string values only")
                    elif len(s) >= MAX_COLUMN_NAME_LENGTH:
                        raise ValueError(
                            f"Column name length in meta features must be shorter than {MAX_COLUMN_NAME_LENGTH}"
                        )
        return v


class DatasetSummary(BaseModel):
    """Describing meta-features of datasets.

    Attributes
    ----------
    columns : dict[str, Column]
        Dictionary of meta features.
        - dict[column_name:str : [dtype:str , meta_features:MetaFeatures | None , has_negative_value:bool]]
    meta_features_pp : MetaFeatures
        Meta-features used to consider preprocessing components.
    meta_features_m : MetaFeatures
        Meta-features used to consider machine learning models.
    has_multi_class_targets : bool
        Whether has multi class targets.
    has_inf_value_targets : bool
        Whether has inf value targets.
    cols_almost_missing_string : list[str] | None, default None
        List of string columns where almost all values are missing.
    cols_almost_missing_numeric : list[str] | None, default None
        List of numerical columns where almost all values are missing.
    cols_str_other : list[str] | None, default None
        List of string columns which are neither categorical nor text columns.
    """

    columns: dict[ColumnName, Column]
    meta_features_pp: MetaFeatures
    meta_features_m: MetaFeatures
    has_multi_class_targets: bool
    has_inf_value_targets: bool
    cols_almost_missing_string: Optional[list[ColumnName]] = None
    cols_almost_missing_numeric: Optional[list[ColumnName]] = None
    cols_str_other: Optional[list[ColumnName]] = None

    @field_validator("columns", "cols_almost_missing_string", "cols_almost_missing_numeric", "cols_str_other")
    def check_num_of_columns(cls, v):
        if v is None:
            return v
        if len(v.keys() if isinstance(v, dict) else v) >= MAX_NUM_OF_COLUMNS:
            raise ValueError(f"The number of columns must be smaller than {MAX_NUM_OF_COLUMNS}")
        return v

    @field_validator("meta_features_pp", "meta_features_m")
    def check_meta_features(cls, v):
        for k, _v in v.items():
            if not re.match(r"feature:[a-z_0-9]+", k):
                raise ValueError(f"'{k}' is invalid as a feature name")
            if isinstance(_v, int):
                pass
            elif isinstance(_v, float):
                pass
            elif isinstance(_v, str) and len(_v) >= MAX_COLUMN_NAME_LENGTH:
                raise ValueError(f"Meta feature string value must be shorter than {MAX_COLUMN_NAME_LENGTH}")
            elif isinstance(_v, list):
                for s in _v:
                    if not isinstance(s, str):
                        raise ValueError("The list must contain string values only")
                    elif len(s) >= MAX_COLUMN_NAME_LENGTH:
                        raise ValueError(
                            f"Column name length in meta features must be shorter than {MAX_COLUMN_NAME_LENGTH}"
                        )
        return v


class ModelLabel(BaseModel):
    """Describing a model name and information to embed it to the code.

    Attributes
    ----------
    label_name: str
        Label name.
    predict_proba : bool, default False
        Whether to predict probability.
    hyperparameters : Any | None, default None
        Hyperparameters.
    meta_features : list[Any], default Field(default_factory=list)
        List of Meta features.
    """

    label_name: String
    predict_proba: bool = False
    hyperparameters: Optional[Any] = None
    meta_features: list[Any] = Field(default_factory=list)

    def __str__(self):
        return self.label_name

    def __repr__(self):
        return str(self)


class SimplePipeline(Code):
    """Generated code with information during adaptation.

    Attributes
    ----------
    pipeline_json : dict, dafault Field(default_factory=lambda: defaultdict(dict))
        Dictionary of pipeline.
    labels : PipelineSkeleton | None , dafault None
        PipelineSkeleton class.
    model : ModelLabel | None , dafault None
        ModelLabel class.
    """

    # pipeline json
    pipeline_json: dict = Field(default_factory=lambda: defaultdict(dict))

    labels: Optional[PipelineSkeleton] = None
    model: Optional[ModelLabel] = None


class Pipeline(SimplePipeline):
    """Generated code with information used internally during adaptation.

    Attributes
    ----------
    task : Task
        Object of the Task class.
    dataset_summary : DatasetSummary
        Object of the DatasetSummary Class.
    config : SapientMLConfig
        Object of the SapientMLConfig.
    adaptation_metric : str | None, default None
        Adaptation metric.
    all_columns_datatypes : dict, default Field(default_factory=dict)
        Dictionary of all column data types.
    inverse_target : bool, default False
        Whether expm1 is required for target columns because log1p is already applied to them.
    sparse_matrix : bool, default False
        Whether the data is converted to sparse matrix in the pipeline.
    train_column_names : list[str], default Field(default_factory=list)
        Column name of training.
    test_column_names : list[str], default Field(default_factory=list)
        Column name of testing.
    is_multi_class_multi_targets : bool, default False
        Whether is multi-class and multi-target.
        - To handle following case;
            - metrics : Accuracy
            - task    : multi-class and multi-targets
        - because sklearn.accuracy_score doesn't support multi-class classification
          involving multiple columns.
    id_columns_for_prediction : list[str], default Field(default_factory=list)
        ID columns for prediction.
    output_dir_path : str, default ""
        Output directory path.
    """

    task: Task
    dataset_summary: DatasetSummary
    config: SapientMLConfig
    adaptation_metric: Optional[String] = None
    all_columns_datatypes: dict = Field(default_factory=dict)
    inverse_target: bool = False
    sparse_matrix: bool = False  # Whether the data is converted to sparse matrix in the pipeline
    train_column_names: list[ColumnName] = Field(default_factory=list)
    test_column_names: list[ColumnName] = Field(default_factory=list)

    # To handle following case;
    #   metrics : Accuracy
    #   task    : multi-class and multi-targets
    # because sklearn.accuracy_score doesn't support multi-class
    is_multi_class_multi_targets: bool = False

    id_columns_for_prediction: list[ColumnName] = Field(default_factory=list)
    output_dir_path: String = ""


def summarize_dataset(df_train: pd.DataFrame, task: Task) -> DatasetSummary:
    """Summarize dataset.

    Parameters
    ----------
        df_train : pd.DataFrame
            Input dataset.
        task : Task
            Object of the Task class.

    Returns
    ----------
        DatasetSummary
    """
    is_multi_classes: list[bool] = []
    for target in task.target_columns:
        is_multi_classes.append(len(df_train[target].unique()) > 1)
    has_multi_class_targets = all(is_multi_classes)

    has_inf_value_targets = bool(df_train[task.target_columns].isin([np.inf, -np.inf]).sum().sum() > 0)

    # handle almost_missing columns
    cols_all_missing = df_train.columns[df_train.isnull().all()]
    cols_almost_missing = df_train.columns[df_train.isna().sum() / len(df_train) > 0.8]
    cols_almost_missing = sorted(list(set(cols_almost_missing) - set(cols_all_missing)))
    cols_almost_missing_numeric = [col for col in cols_almost_missing if is_numeric_dtype(df_train[col])]
    cols_almost_missing_string = [col for col in cols_almost_missing if not is_numeric_dtype(df_train[col])]

    columns = dict()
    for column_name in df_train.columns:
        meta_features = generate_column_meta_features(df_train[[column_name]])

        columns[column_name] = Column(
            dtype=str(df_train[column_name].dtype),
            meta_features=meta_features,
            has_negative_value=bool(
                is_numeric_dtype(df_train[column_name]) and (df_train[[column_name]].values <= 0).any()
            ),
        )

    # Generate the meta-features
    meta_features_pp = generate_pp_meta_features(df_train, task.target_columns)
    if ps_macros.STR_OTHER in meta_features_pp:
        cols_str_other = meta_features_pp[ps_macros.STR_OTHER]
        del meta_features_pp[ps_macros.STR_OTHER]

    is_clf_task = 1 if task.task_type == "classification" else 0
    meta_features_m = generate_model_meta_features(df_train, task.target_columns, is_clf_task)

    return DatasetSummary(
        columns=columns,
        meta_features_pp=meta_features_pp,
        meta_features_m=meta_features_m,
        has_multi_class_targets=has_multi_class_targets,
        has_inf_value_targets=has_inf_value_targets,
        cols_almost_missing_string=cols_almost_missing_string,
        cols_almost_missing_numeric=cols_almost_missing_numeric,
        cols_str_other=cols_str_other,
    )



# Copyright 2023-2024 The SapientML Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import logging
import os
import shutil
import subprocess
import time
from multiprocessing import Pool, cpu_count

from sapientml.util.logging import setup_logger
from sapientml_core import internal_path
from sapientml_core.training import project_corpus
from sapientml_core.training.augmentation import mutator
from tqdm import tqdm

TIME_LIMIT = "10m"
SKIP_GOOD_RUN = True
SKIP_EXECED_ORIGINAL = False
SKIP_TLE = False
REGEN_MUTATION = True
GEN_DF_DUMPER = True

# whether to investigate notebooks other than the best notebook
run_all_notebooks = False

run_original = False
run_all_types = True
run_df_dumper = False

logger = setup_logger()


def execute(
    pipeline_name_without_extension,
    modelname,
    script_path,
    work_dir=None,
    skip_good_run=True,
    skip_timeout=True,
):
    """Executes the pipelines and update the results.

    Parameters
    ----------
    pipeline_name_without_extension : str
        It returns a string variable
    modelname : str
        It returns a string variable
    script_path : PosixPath
        It contains path from pathlib.PosixPath
    work_dir : PosixPath
        It contains path from pathlib.PosixPath
    skip_good_run : bool
        True and otherwise False
    skip_timeout : bool
        True and otherwise False

    """
    run_output_dir = internal_path.execution_cache_dir / pipeline_name_without_extension / modelname
    if not os.path.exists(run_output_dir):
        os.makedirs(run_output_dir)

    if os.path.exists(run_output_dir / "returncode.txt"):
        with open(run_output_dir / "returncode.txt", "r", encoding="utf-8") as f:
            code = f.read()
        code = int(code.strip())
        if code == 0 and skip_good_run:
            with open(os.path.join(run_output_dir, "stdout.txt"), "r", encoding="utf-8") as f:
                stdout = f.read()
            stdout = stdout.strip()
            stdout = stdout if len(stdout) <= 1000 else stdout[-1000:]
            logger.info(stdout)
            logger.info("a good run, skip")
            return
        if code == 124 and skip_timeout:
            logger.info("previously timeout, skip")
            return

    stime = time.time()

    subproc = subprocess.Popen(
        f"timeout {TIME_LIMIT} python " + str(script_path).replace(" ", "\\ "),
        cwd=work_dir,
        shell=True,
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
        text=True,
    )

    stdout, stderr = subproc.communicate()
    ttime = time.time()
    if not os.path.exists(run_output_dir):
        os.makedirs(run_output_dir)
    with open(os.path.join(run_output_dir, "stdout.txt"), "w", encoding="utf-8") as f:
        f.write(stdout)
    with open(os.path.join(run_output_dir, "stderr.txt"), "w", encoding="utf-8") as f:
        f.write(stderr)
    with open(os.path.join(run_output_dir, "returncode.txt"), "w", encoding="utf-8") as f:
        f.write(str(subproc.returncode))
    with open(os.path.join(run_output_dir, "time.txt"), "w", encoding="utf-8") as f:
        f.write(str(ttime - stime))
    if subproc.returncode != 0:
        logger.warning(
            f"!!! Error encountered @ {pipeline_name_without_extension}/{modelname}: returncode = {subproc.returncode}"
        )

    else:
        stdout = stdout.strip()
        stdout = stdout if len(stdout) <= 1000 else stdout[-1000:]


def run(args):
    """Wrapper function for execute() function.

    Parameters
    ----------
    args : tuple

    Returns
    -------
    j : int
    now_t : str

    """
    j, now_t, pipeline_name_without_extension = args
    execute(
        pipeline_name_without_extension,
        now_t,
        internal_path.execution_cache_dir / pipeline_name_without_extension / now_t / "script.py",
        internal_path.clean_dir,
        SKIP_GOOD_RUN,
    )
    return j, now_t


def generate_mutated_accuracy(project):
    """Executes original and mutated pipelines.

    Parameters
    ----------
    project : ProjectInfo
        It contains ProjectInfo class details

    """
    pipeline_name_without_extension = project.notebook_name

    with open(project.pipeline_path, "r", encoding="utf-8") as f:
        original_script = f.read()

    original_script = original_script.replace("../../dataset", str(internal_path.corpus_path / "dataset"))

    script_path = internal_path.execution_cache_dir / pipeline_name_without_extension / "original"
    if not os.path.exists(script_path):
        os.makedirs(script_path)
    script_path = script_path / "script.py"
    with open(script_path, "w", encoding="utf-8") as f:
        f.write(original_script)

    # Running Original Pipeline
    execute(
        pipeline_name_without_extension,
        "original",
        script_path,
        internal_path.clean_dir,
        SKIP_GOOD_RUN,
    )

    origin_type, muted_types = "", []
    try:
        origin_type, muted_types = mutator.model_mutation(
            script_path,
            internal_path.execution_cache_dir / pipeline_name_without_extension,
            REGEN_MUTATION,
            GEN_DF_DUMPER,
        )
    except Exception:
        pass

    if run_all_types:
        exec_trace_files = ["stdout.txt", "stderr.txt", "time.txt", "returncode.txt"]
        if SKIP_EXECED_ORIGINAL and all(
            [
                os.path.exists(internal_path.execution_cache_dir / pipeline_name_without_extension / "original" / item)
                for item in exec_trace_files
            ]
        ):
            # detected that the original is already executed and allow to skip
            for file in exec_trace_files:
                shutil.copy(
                    internal_path.execution_cache_dir / pipeline_name_without_extension / "original" / file,
                    internal_path.execution_cache_dir / pipeline_name_without_extension / origin_type / file,
                )
            pass
        else:
            # run original
            script_path = (
                internal_path.execution_cache_dir / pipeline_name_without_extension / origin_type / "script.py"
            )
            if os.path.exists(script_path):
                execute(
                    pipeline_name_without_extension,
                    origin_type,
                    script_path,
                    internal_path.clean_dir,
                    SKIP_GOOD_RUN,
                )

        muted_types = [(j, now_t, pipeline_name_without_extension) for j, now_t in enumerate(muted_types)]

        # then run muted types
        with Pool(cpu_count() // 4 or 1) as p:
            for j, now_t in tqdm(p.imap(run, muted_types)):
                logger.info(f"  [{j + 1}/{len(muted_types) + 1}] run type {now_t}")
        p.join()


def main(n, k):
    """Creates augmented pipelines and executes it.

    Parameters
    ----------
    n : int
    k : int

    """
    corpus = project_corpus.ProjectCorpus()
    if not os.path.exists(internal_path.execution_cache_dir):
        os.makedirs(internal_path.execution_cache_dir)
    logging.basicConfig(
        filename=internal_path.execution_cache_dir / f"mutation_{k}.log",
        filemode="a",
        format="%(asctime)s,%(msecs)d %(name)s %(levelname)s %(message)s",
        datefmt="%H:%M:%S",
    )
    for i, project in enumerate(corpus.project_list):
        if i % n != k:
            continue
        logger.info(f"{i+1}: Running mutation for: {project.notebook_name}")

        try:
            generate_mutated_accuracy(project)
        except Exception:
            import traceback

            logger.warning(traceback.print_exc())

    with open(internal_path.execution_cache_dir / f"mutation_{k}.finished", "w") as f:
        f.write("finished")


if __name__ == "__main__":
    import sys

    args = sys.argv[1:]
    n = int(args[0])
    k = int(args[1])

    if len(args) > 2:
        internal_path.training_cache = internal_path.training_cache / args[2]
        internal_path.execution_cache_dir = internal_path.training_cache / "exec_info"
    main(n, k)



# Copyright 2023-2024 The SapientML Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import re
from pathlib import Path
from typing import Tuple

import fasttext
import ipadic
import MeCab
import numpy as np
import pandas as pd
import requests
from jinja2 import Environment, FileSystemLoader
from sapientml.generator import CodeBlockGenerator
from sapientml.params import Code, Dataset, Task
from sapientml.util.logging import setup_logger

from .params import DefaultPreprocessConfig

logger = setup_logger()

INHIBITED_SYMBOL_PATTERN = re.compile(r"[\{\}\[\]\",:<'\\]+")


template_env = Environment(loader=FileSystemLoader(f"{os.path.dirname(__file__)}/templates"), trim_blocks=True)


def _is_strnum_column(c):
    c2 = c.loc[c.notnull()]
    c2 = pd.to_numeric(c2, errors="coerce")
    ratio = c2.notnull().sum() / c2.shape[0]
    return ratio > 0.9


def _is_category_column(c):
    c2 = c.loc[c.notnull()]
    num = c2.nunique()
    if num <= 2:
        return "binary"
    elif num <= 20 and (num / c2.shape[0] < 0.3):
        return "small"
    elif c2.value_counts(normalize=True).head(20).sum() > 0.8:
        return "large"
    else:
        return False


def _confirm_mixed_type(df: pd.DataFrame) -> list[str]:
    mixed_df_cols = []
    for df_col in df.columns:
        df_per_col = df[df_col].replace([-np.inf, np.inf], np.nan)
        df_per_col2 = df_per_col[df_per_col.notnull()]
        types_per_col = pd.api.types.infer_dtype(df_per_col2)

        is_strnum = _is_strnum_column(df[df_col])

        if types_per_col == "mixed" or types_per_col == "mixed-integer" or (types_per_col == "string" and is_strnum):
            mixed_df_cols.append(df_col)

    return mixed_df_cols


def _render(tpl, *args, **kwargs):
    code = tpl.render(*args, **kwargs)
    return "\n".join([line for line in code.split("\n") if len(line) > 0]) + "\n\n"


def check_column_language(df: pd.DataFrame) -> list[str]:
    """This function checks the language of column using fasttext.

    Parameters
    ----------
    df : pd.DataFrame

    Returns
    -------
    col_language_dict : list[str]
        Return list of all the column of specific language.

    """
    os.makedirs(Path(os.path.dirname(__file__)) / "lib", exist_ok=True)
    model_path = Path(os.path.dirname(__file__)) / "lib" / "lid.176.bin"
    if not model_path.exists():
        response = requests.get("https://dl.fbaipublicfiles.com/fasttext/supervised-models/lid.176.bin")
        response.raise_for_status()  # check download status and raise if download is failed
        with open(model_path, mode="wb") as f:
            f.write(response.content)
    fasttext.FastText.eprint = lambda x: None
    model = fasttext.load_model(model_path._str)
    object_columns = df.select_dtypes("object").columns
    col_language_dict = {}
    for col in object_columns:
        if not _is_category_column(df[col]):
            c2 = df[col][df[col].notnull()]
            if c2.shape[0] > 1000:
                c2 = c2.sample(1000, random_state=17)
            value_languages = (
                c2.astype(str)
                .str.replace("\n", "")
                .str.replace("\r", "")
                .apply(lambda x: model.predict(x)[0][0].split("__label__")[1])
            )
            col_language_dict[col] = value_languages.mode()[0]
    return col_language_dict


def tokenize(text, use_pos_list, use_word_stemming, tokenizer):
    """tokenize.

    Parameters
    ----------
    text : str
    use_pos_list : list[str]
       List of parts-of-speech to be used during text analysis.
       This variable is used for japanese texts analysis.
       Select the part of speech below.
       "名詞", "動詞", "形容詞", "形容動詞", "副詞"
    use_word_stemming : bool
       Specify whether or not word stemming is used.
       This variable is used for japanese texts analysis.
    tokenizer : MeCab tokenizer

    Returns
    -------
    terms : list[str]

    """
    node = tokenizer.parseToNode(text)
    terms = []
    while node:
        features = node.feature.split(",")
        pos = features[0]
        if pos != "BOS/EOS":
            if use_word_stemming:
                term = features[6]
                if (pos == "名詞") & (features[1] == "数"):
                    term = node.surface
            else:
                term = node.surface
            if use_pos_list:
                if pos in use_pos_list:
                    terms.append(term)
            else:
                terms.append(term)
        node = node.next
    return " ".join(terms)


def check_cols_has_symbols(columns: list) -> list[str]:
    """check_cols_has_symbols.

    Parameters
    ----------
    columns : list

    Returns
    -------
    cols_has_symbols : list[str]
        Return the list of column that has symbols.

    """
    cols_has_symbols = []
    for col in columns:
        if INHIBITED_SYMBOL_PATTERN.search(col):
            cols_has_symbols.append(col)
    return cols_has_symbols


def remove_symbols(column_name: str) -> str:
    """remove_symbols.

    Parameters
    ----------
    column_name : str

    Returns
    -------
    column_name : str
        Return coulum name after removing symbols.

    """
    return INHIBITED_SYMBOL_PATTERN.sub("", column_name)


class DefaultPreprocess(CodeBlockGenerator):
    def __init__(self, **kwargs):
        self.config = DefaultPreprocessConfig(**kwargs)

    def generate_code(self, dataset: Dataset, task: Task) -> Tuple[Dataset, Code]:
        """Generate code of rule-based preprocessing.

        This function will update dataset by removing special symbols, mixed-type column and
        replacing 'inf' with 'nan' for meta feature calculation and handling Japanese text columns
        using MeCab libarary.

        Parameters
        ----------
        dataset : Dataset
           Object of Dataset class containing the details of dataset.
        task : Task
           Object of Task class containing the details of task.

        Returns
        -------
        dataset: Dataset
            Dataset after preprocessing.
        code: Code
            Code of the preprocessing.
        """
        code = Code()
        df = pd.concat(dataset.get_dataframes()).reset_index(drop=True)

        # Remove special symbols that interfere with visualization and model training
        cols_has_symbols = []
        cols_has_symbols = check_cols_has_symbols(df.columns.to_list())
        if cols_has_symbols:
            logger.warning(
                f"Symbols that inhibit training and visualization will be removed from column name {str(cols_has_symbols)}."
            )
            df = df.rename(columns=lambda col: remove_symbols(col) if col in cols_has_symbols else col)
            task.target_columns = [
                remove_symbols(col) if col in cols_has_symbols else col for col in task.target_columns
            ]
            tpl = template_env.get_template("rename_columns.py.jinja")
            code.validation += _render(tpl, training=True, test=True, cols_has_symbols=cols_has_symbols)
            code.test += _render(tpl, training=True, test=True, cols_has_symbols=cols_has_symbols)
            code.train += _render(tpl, training=True, test=False, cols_has_symbols=cols_has_symbols)
            code.predict += _render(tpl, training=False, test=True, cols_has_symbols=cols_has_symbols)

        # If None is intentionally inserted in the data, an error occurs, so we have added an action to change None to "np.nan."
        if df.isin([None]).any(axis=None):
            df = df.replace([None], np.nan)
            tpl = template_env.get_template("none_has_columns.py.jinja")
            code.validation += _render(tpl, training=True, test=True)
            code.test += _render(tpl, training=True, test=True)
            code.train += _render(tpl, training=True, test=False)
            code.predict += _render(tpl, training=False, test=True)

        # handle list(tuple, dict) value in dataframe.
        # in generated scripts, visualisation will be executed before pre-processing such as handle mixed-type.
        # so, need to check before mixed-type column names are added to suppress errors during visualisation.
        cols_iterable_values = []
        for col in df.columns:
            exist_list_values = [x for x in df[col] if type(x) in [list, tuple, dict]]
            if len(exist_list_values) > 0:
                cols_iterable_values.append(col)
                df[col] = df[col].fillna("").astype(str)
        if cols_iterable_values:
            tpl = template_env.get_template("handle_iterable_values.py.jinja")
            code.validation += _render(tpl, training=True, test=True, cols_iterable_values=cols_iterable_values)
            code.test += _render(tpl, training=True, test=True, cols_iterable_values=cols_iterable_values)
            code.train += _render(tpl, training=True, test=False, cols_iterable_values=cols_iterable_values)
            code.predict += _render(tpl, training=False, test=True, cols_iterable_values=cols_iterable_values)

        # handle mixed-type columns
        # split a columns into 2 columns, one column has only numeric, another columns has only string
        # this operation should be done before calculating meta features
        mix_typed_cols = _confirm_mixed_type(df.drop(task.target_columns, axis=1))
        cols_numeric_and_string = []
        for col in mix_typed_cols:
            cols_numeric_and_string.append(col)
            only_str = col + "__str"
            only_num = col + "__num"
            df[only_str] = np.where(pd.to_numeric(df[col], errors="coerce").isnull(), df[col], np.nan)
            df[only_str] = np.where(df[only_str].notnull(), df[only_str].astype(str), np.nan)
            # without .astype(float), cannot recongnize as `int` or `float`, leading to generate inappropriate code snippet
            df[only_num] = np.where(pd.to_numeric(df[col], errors="coerce").isnull(), np.nan, df[col]).astype(float)
            df = df.drop(col, axis=1)
        if cols_numeric_and_string:
            tpl = template_env.get_template("handle_mixed_typed_columns.py.jinja")
            code.validation += _render(tpl, training=True, test=True, cols_numeric_and_string=cols_numeric_and_string)
            code.test += _render(tpl, training=True, test=True, cols_numeric_and_string=cols_numeric_and_string)
            code.train += _render(tpl, training=True, test=False, cols_numeric_and_string=cols_numeric_and_string)
            code.predict += _render(tpl, training=False, test=True, cols_numeric_and_string=cols_numeric_and_string)

        # meta features must be calculated after replacing inf with nan,
        # becuase the replaced nan must be preprocessed in the generated code.
        # handling inf as nan must be AFTER to_csv() to keep dataset files intact
        inf_bool_s = df.drop(task.target_columns, axis=1).isin([np.inf, -np.inf]).any()
        cols_inf_values = inf_bool_s.index[inf_bool_s].tolist()
        if cols_inf_values:
            df[cols_inf_values] = df[cols_inf_values].replace([np.inf, -np.inf], np.nan)
            tpl = template_env.get_template("handle_inf_columns.py.jinja")
            code.validation += _render(tpl, training=True, test=True, cols_inf_values=cols_inf_values)
            code.test += _render(tpl, training=True, test=True, cols_inf_values=cols_inf_values)
            code.train += _render(tpl, training=True, test=False, cols_inf_values=cols_inf_values)
            code.predict += _render(tpl, training=False, test=True, cols_inf_values=cols_inf_values)

        # handle Japanese text columns
        cols_japanese_text = []
        column_language_dict = check_column_language(df.drop(task.target_columns, axis=1))
        cols_japanese_text = [key for key, val in column_language_dict.items() if val == "ja"]
        if cols_japanese_text:
            tokenizer = MeCab.Tagger(ipadic.MECAB_ARGS)
            for col in cols_japanese_text:
                df[col] = df[col].fillna("")
                df[col] = df[col].apply(
                    lambda x: tokenize(x, self.config.use_pos_list, self.config.use_word_stemming, tokenizer)
                )
            tpl = template_env.get_template("handle_japanese_text.py.jinja")
            code.validation += _render(
                tpl, config=self.config, training=True, test=True, cols_japanese_text=cols_japanese_text
            )
            code.test += _render(
                tpl, config=self.config, training=True, test=True, cols_japanese_text=cols_japanese_text
            )
            code.train += _render(
                tpl, config=self.config, training=True, test=False, cols_japanese_text=cols_japanese_text
            )
            code.predict += _render(
                tpl, config=self.config, training=False, test=True, cols_japanese_text=cols_japanese_text
            )

        cols_one_value_only = df.columns[df.nunique(dropna=False) == 1].tolist()
        if cols_one_value_only:
            df = df.drop(cols_one_value_only, axis=1)
            tpl = template_env.get_template("drop_one_value_columns.py.jinja")
            code.validation += _render(tpl, training=True, test=True, cols_one_value_only=cols_one_value_only)
            code.test += _render(tpl, training=True, test=True, cols_one_value_only=cols_one_value_only)
            code.train += _render(tpl, training=True, test=False, cols_one_value_only=cols_one_value_only)
            code.predict += _render(tpl, training=False, test=True, cols_one_value_only=cols_one_value_only)

        dataset.training_dataframe = df

        return dataset, code



from decimal import ROUND_HALF_UP, Decimal

import pandas as pd
from sklearn.model_selection import train_test_split


def _sampled_training(dev_training_dataset, train_size, stratify, task_type) -> pd.DataFrame:
    sampled_training_dataset, _ = train_test_split(
        dev_training_dataset,
        train_size=train_size,
        stratify=stratify if task_type == "classification" else None,
    )
    return sampled_training_dataset  # type: ignore


def sample_dataset(
    dataframe: pd.DataFrame,
    sample_size: int,
    target_columns: list[str],
    task_type: str,
) -> pd.DataFrame:
    # Sample the training set if the dataset is big
    # FIXME
    sampled_training_dataset = None
    num_of_rows = len(dataframe.index)
    if num_of_rows >= sample_size:
        rare_labels = []
        dataframe_alltargets = None
        if task_type == "classification":
            dataframe_alltargets = dataframe[target_columns].astype(str).apply("".join, axis=1)
            label_count = dataframe_alltargets.value_counts()
            rare_labels = label_count.loc[label_count == 1].index.tolist()

        if rare_labels and dataframe_alltargets is not None:
            dataframe_rare = dataframe[dataframe_alltargets.isin(rare_labels)]
            rare_index = dataframe_rare.index.values

            dataframe_wo_rare = dataframe.drop(rare_index)

            num_of_labels = [len(dataframe_wo_rare[target].value_counts()) for target in target_columns]

            rare_to_all_ratio = int(
                Decimal(sample_size * len(dataframe_rare) / len(dataframe)).quantize(
                    Decimal("0"), rounding=ROUND_HALF_UP
                )
            )
            not_rare_to_all_ratio = int(
                Decimal(sample_size * len(dataframe_wo_rare) / len(dataframe)).quantize(
                    Decimal("0"), rounding=ROUND_HALF_UP
                )
            )

            stratify_wo_rare = None

            if len(dataframe_rare) == len(dataframe):
                sampled_training_dataset = _sampled_training(dataframe, sample_size, None, task_type)

            elif rare_to_all_ratio in [0, 1]:
                sampled_training_dataset_rare = dataframe_rare

                if max(num_of_labels) >= sample_size:
                    stratify_wo_rare = None
                else:
                    stratify_wo_rare = dataframe_wo_rare[target_columns]
                sampled_training_dataset_wo_rare = _sampled_training(
                    dataframe_wo_rare,
                    sample_size - len(sampled_training_dataset_rare),
                    stratify_wo_rare,
                    task_type,
                )

                sampled_training_dataset = pd.concat(
                    [sampled_training_dataset_wo_rare, sampled_training_dataset_rare]  # type: ignore
                )

            elif not_rare_to_all_ratio in [0, 1]:
                sampled_training_dataset_wo_rare = dataframe_wo_rare
                sampled_training_dataset_rare = _sampled_training(
                    dataframe_rare,
                    sample_size - len(sampled_training_dataset_wo_rare),
                    None,
                    task_type,
                )

                sampled_training_dataset = pd.concat(
                    [sampled_training_dataset_wo_rare, sampled_training_dataset_rare]  # type: ignore
                )

            else:
                if max(num_of_labels) >= sample_size:
                    stratify_wo_rare = None
                else:
                    stratify_wo_rare = dataframe_wo_rare[target_columns]

                sampled_training_dataset_wo_rare = _sampled_training(
                    dataframe_wo_rare, not_rare_to_all_ratio, stratify_wo_rare, task_type
                )
                sampled_training_dataset_rare = _sampled_training(dataframe_rare, rare_to_all_ratio, None, task_type)

                sampled_training_dataset = pd.concat(
                    [sampled_training_dataset_wo_rare, sampled_training_dataset_rare]  # type: ignore
                )

        else:
            num_of_labels = [len(dataframe[target].value_counts()) for target in target_columns]
            if max(num_of_labels) >= sample_size:
                stratify_wo_rare = None
            else:
                stratify_wo_rare = dataframe[target_columns]

            sampled_training_dataset = _sampled_training(dataframe, sample_size, stratify_wo_rare, task_type)
        return sampled_training_dataset
    else:
        return dataframe



# Copyright 2023-2024 The SapientML Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import json
from pathlib import Path

import pandas as pd
from sapientml.util.logging import setup_logger
from sapientml_core import internal_path
from sapientml_core.meta_features import (
    collect_labels,
    compute_model_meta_features,
    compute_pp_meta_features,
    real_feature_preprocess,
)
from sapientml_core.training.project_corpus import ProjectCorpus
from sapientml_core.util import file_util
from tqdm import tqdm

logger = setup_logger()


# Get the stored features summary
def get_feature_usage_summary():
    """Open and load feature_analysis_summary.json file.

    Returns
    -------
    current_summary : dict
        Loaded summary file
    """
    try:
        with open(internal_path.training_cache / "feature_analysis_summary.json", "r", encoding="utf-8") as f:
            current_summary = json.load(f)
    except Exception:
        current_summary = {}
    return current_summary


def get_column_usage_summary_in_pipeline(pipeline_path, feature_usage_summary):
    """Extract used_cols in pipeline.

    Parameters
    ----------
    pipeline_path :
        pipeline path
    feature_usage_summary : dict
        Summary data

    Returns
    -------
    None
        If pipeline_path not in feature_usage_summary.
    used_columns : pd.Series
        Extractetd columns
    """
    if pipeline_path in feature_usage_summary:
        pipeline_summary = feature_usage_summary[pipeline_path]
    else:
        return None

    if "status" in pipeline_summary and pipeline_summary["status"] == "FINALIZED":
        used_columns = pipeline_summary["used_cols"]
    else:
        used_columns = None

    return used_columns


def get_dataset_folder_name(file_path):
    notebook_info_path = file_path.replace(".py", ".info.json")
    dataset_folder_name = ""
    try:
        with open(notebook_info_path, "r", encoding="utf-8") as notebook_info_file:
            notebook_info = json.load(notebook_info_file)
            if isinstance(notebook_info, list):
                notebook_info = notebook_info[1]

            if isinstance(notebook_info, dict):
                dataset_folder_name = notebook_info["dataset_folder"]
    except Exception:
        logger.warning("Could not read JSON info file: {}".format(notebook_info_path))
    return dataset_folder_name


def collect_training_meta_feature(mode):
    """Read csv and Generate the meta-features.

    Parameters
    ----------
    mode : str
        "clean" or "as-is"

    Returns
    -------
    final_pp_meta_features : pd.Dataframe
        Dataframe of pp_meta_features and labels
    final_model_meta_feature_df : pd.Dataframe
        Dataframe of model_meta_feature and models
    """
    pp_meta_features = []
    model_meta_features = []
    feature_usage_summary = get_feature_usage_summary()

    corpus = ProjectCorpus()
    projects = corpus.project_list

    total_number_target_pipelines = len(projects)

    column_not_found = 0

    for i in tqdm(range(0, total_number_target_pipelines)):
        project = projects[i]
        pipeline_path = project.pipeline_path
        file_name = projects[i].file_name

        dataset_folder_name = get_dataset_folder_name(pipeline_path)
        project.csv_name = dataset_folder_name + "_" + project.csv_name

        dataset_path = project.dataset_path
        logger.debug(
            "EXTRACTING META-FEATURES:", i + 1, "out of ", total_number_target_pipelines, "PIPELINE:", pipeline_path
        )
        try:
            df = file_util.read_csv(
                Path(project.dataset_path),
                Path(project.pipeline_path),
            )
            if mode != "as-is":
                import copy

                df_copy = copy.deepcopy(df)
                used_cols = get_column_usage_summary_in_pipeline(file_name, feature_usage_summary)
                if used_cols is not None:
                    df = df[used_cols]
        except KeyError:
            column_not_found += 1
            df = df_copy
        except Exception:
            logger.warning("Could not read CSV: {}".format(dataset_path))
            continue

        df.rename(columns=lambda x: x.strip(), inplace=True)

        proj_name = pipeline_path
        target_column_name = projects[i].target_column_name
        # Generate the meta-features
        pp_meta_feature_dict = compute_pp_meta_features(df, proj_name, project, target_column_name)
        model_meta_feature_dict = compute_model_meta_features(df, proj_name, project, target_column_name)

        if pp_meta_feature_dict is not None:
            pp_meta_features.append(pp_meta_feature_dict)

        if model_meta_feature_dict is not None:
            model_meta_features.append(model_meta_feature_dict)

    pp_meta_features_df = pd.DataFrame(pp_meta_features)
    model_meta_feature_df = real_feature_preprocess(pd.DataFrame(model_meta_features))

    # Add the labels
    annotated_notebooks_path = internal_path.project_labels_path
    labels_df = collect_labels(annotated_notebooks_path)

    # Merge two dataframes
    final_pp_meta_features = pd.merge(pp_meta_features_df, labels_df, on="file_name", how="inner")
    logger.debug(f"final_pp_meta_features shape:{final_pp_meta_features.shape}")

    models = []
    is_clf = []

    for _, row in labels_df.iterrows():
        model_added = False
        for col in labels_df.columns.to_list():
            if isinstance(col, str) and col.startswith("MODEL:") and row[col] == 1:
                models.append(col)
                # 0: Regresssor 1: Classifier
                clf = 0 if ":Regressor:" in col else 1
                is_clf.append(clf)
                model_added = True
                break
        if not model_added:
            models.append("MODEL:NOT_DETECTED")
            is_clf.append(2)

    labels_df["target"] = models

    model_df = labels_df[["file_name", "target"]]
    model_df["feature:is_clf"] = is_clf
    model_df["weight"] = 1
    # Merge two dataframes
    final_model_meta_feature_df = pd.merge(model_meta_feature_df, model_df, on="file_name", how="inner")
    logger.debug(f"final_model_meta_feature_df shape:{final_model_meta_feature_df.shape}")

    return final_pp_meta_features, final_model_meta_feature_df


def main():
    """
    This main function calls the subfunctions and outputs the result to csv.

    Description of mode : "clean" | "as-is"
        as-is mode: compute meta-features based on all the meta-features in the dataset
        clean mode: only use the meta-features that are used in the pipeline

    """
    mode = "clean"
    final_pp_meta_features_df, final_model_meta_feature_df = collect_training_meta_feature(mode)
    pp_meta_features_path = internal_path.training_cache / "pp_metafeatures_training.csv"
    model_meta_features_path = internal_path.training_cache / "model_metafeatures_training.csv"
    final_pp_meta_features_df.to_csv(pp_meta_features_path, index=False)
    final_model_meta_feature_df.to_csv(model_meta_features_path, index=False)
    logger.info(f"Meta-data successsfully stored at:{internal_path.training_cache}")


if __name__ == "__main__":
    import argparse

    parser = argparse.ArgumentParser()
    parser.add_argument("--tag", type=str, help="Tag for output files and dirs.")
    args = parser.parse_args()
    if args.tag:
        internal_path.training_cache = internal_path.training_cache / args.tag
    main()



# Copyright 2023-2024 The SapientML Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import ast
import os
import subprocess
import time
from multiprocessing import Pool, cpu_count
from pathlib import Path

from sapientml.util.logging import setup_logger
from sapientml_core import internal_path
from sapientml_core.training import project_corpus

logger = setup_logger()


def catch_names(node):
    if isinstance(node, list):
        ret = set()
        for n in node:
            ret = ret.union(catch_names(n))
        return ret
    elif isinstance(node, ast.AST):
        ret = set()
        for subnode in ast.walk(node):
            if isinstance(subnode, ast.Name):
                ret.add(subnode)
        return ret
    else:
        return {}


def catch_first_name(assign_tgt):
    if isinstance(assign_tgt, ast.Name):
        return assign_tgt
    elif isinstance(assign_tgt, ast.Attribute):
        return catch_first_name(assign_tgt.value)
    elif isinstance(assign_tgt, ast.Subscript):
        real_target = assign_tgt
        while isinstance(real_target, ast.Subscript):
            real_target = real_target.value
        return catch_first_name(real_target)
    elif isinstance(assign_tgt, ast.Tuple) or isinstance(assign_tgt, ast.List):
        return catch_first_name(assign_tgt.elts[0])
    else:
        raise Exception(f"Unsupport type {type(assign_tgt)}")


def get_method_name(call_node):
    if isinstance(call_node, ast.Call) and isinstance(call_node.func, ast.Attribute):
        return call_node.func.attr
    else:
        return ""


class Injector(ast.NodeTransformer):
    """
    Instruments the pipelines with code snippets to collect snapshots of dataset.
    """

    def __init__(self):
        self.training_stmt_pattern = ["Dataset", "fit", "OLS", "DMatrix"]
        self.trace_vars = set()
        self.last_train_lineno = None

    def run(self, source_code_after_fixing_csv_path, script_path, new_script_path):
        tree = ast.parse(source=source_code_after_fixing_csv_path, filename=script_path)

        tree = self.visit(tree)

        loading_statements = ast.parse(
            f"""
with open("{new_script_path.replace('.py', '.txt')}", "w") as f:
    f.write('started')
import types
import importlib.machinery
loader = importlib.machinery.SourceFileLoader('df_collector', '{str(Path(__file__).parent / "df_collector.py")}')
mod = types.ModuleType(loader.name)
loader.exec_module(mod)
update_column_names = mod.update_column_names
collector = dict()
"""
        ).body

        # dumping statement
        # also dump the line no. of our determined training stmt
        dumping_statement = ast.parse(
            f"""
collector = [collector, {self.last_train_lineno}]
import json
with open("{new_script_path.replace('.py', '.json')}", 'w', encoding='utf-8') as f: json.dump(collector, f, indent=4)
"""
        ).body

        tree.body = loading_statements + tree.body + dumping_statement

        return ast.unparse(tree)

    def visit_Call(self, node):
        if self.last_train_lineno is None and get_method_name(node) in self.training_stmt_pattern:
            self.last_train_lineno = node.lineno
        return node

    def visit_Assign(self, node):
        if self.last_train_lineno and node.lineno >= self.last_train_lineno:
            return node
        target = catch_first_name(node.targets[0])
        var_out = catch_names(node.targets)
        var_in = catch_names(node.value)

        # debug
        if isinstance(target, ast.Subscript):
            logger.debug(node.lineno)
            logger.debug(ast.dump(node.targets[0]))

        target = target.id
        var_out = set([item.id for item in var_out if isinstance(item, ast.Name)])
        var_in = set([item.id for item in var_in if isinstance(item, ast.Name)])

        if get_method_name(node.value) == "read_csv":
            self.trace_vars.add(target)

        if any([item in self.trace_vars for item in var_in]):
            for item in var_out:
                self.trace_vars.add(item)

        if target in self.trace_vars:
            return [
                node,
                ast.parse(f"update_column_names(collector, {node.lineno}, {target}, '{target}')"),
            ]
        return node


def instrument_pipeline(pipeline, new_path):
    """Instruments a given pipeline with code snippets to collect snapshots of dataset using Injector.

    Parameters
    ----------
    pipeline : str
        pipeline path
    new_path : str
        new_path of dataset
    """
    with open(pipeline, "r", encoding="utf-8") as f:
        source_code = f.read()
    source_code_after_fixing_csv_path = source_code.replace("../../dataset", str(internal_path.corpus_path / "dataset"))
    source_code_after_injecting_targeted_lines = Injector().run(source_code_after_fixing_csv_path, pipeline, new_path)
    os.makedirs(os.path.dirname(new_path), exist_ok=True)
    with open(new_path, "w", encoding="utf-8") as f:
        f.write(source_code_after_injecting_targeted_lines)


def run(file_path):
    """Executes the instrumented version of the pipeline to store the snapshots of the dataset.

    Parameters
    ----------
    file_path : str

    Returns
    -------
    int
       proccess of returncode

    """
    process = subprocess.Popen(
        "python" + " " + file_path,
        shell=True,
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
    )
    output, err = process.communicate()
    output_content = output.decode("utf-8")
    error = err.decode("utf-8")
    logger.info(f"output_content:{output_content}")
    logger.error(f"error:{error}")
    return process.returncode


def dump_final_dataframe(project):
    """This function Validating  dataset_pkl  and replacing .py file with col_dumper file.

    Parameters
    ----------
    project : ProjectInfo
        The parameter has details information of pipleline.

    Returns
    -------
    None
       if file_path not exit.

    Raises
    ------
    AssertionError
        This will give error when assert returncode == 0.

    """
    pipeline_path = project.pipeline_path
    project_root = internal_path.clean_dir
    pipeline_relative_path = project.pipeline_path[len(str(project_root)) + 1 :]
    new_script_path = str(internal_path.training_cache / "dataset-snapshots" / pipeline_relative_path).replace(
        ".py", "_col_dumper.py"
    )
    if os.path.exists(new_script_path.replace(".py", ".json")):
        return None
    instrument_pipeline(pipeline_path, new_script_path)

    dataset_pkl = new_script_path.replace(".py", ".pkl")
    if os.path.exists(dataset_pkl):
        return None

    stime = time.time()
    returncode = run(new_script_path)
    ttime = time.time()
    logger.info(f"TIME SPENT:, {ttime - stime}")
    assert returncode == 0


def _run(data):
    total_number_target_pipelines, projects, i = data
    logger.info(f"RUNNING:, {i + 1}, out of, {total_number_target_pipelines}, PIPELINE: {projects[i].pipeline_path}")
    try:
        dump_final_dataframe(projects[i])
    except Exception:
        import traceback

        traceback.print_exc()


def main():
    """This main function getting all the pipeline details and parse the pipelines."""
    corpus = project_corpus.ProjectCorpus()
    projects = corpus.project_list

    total_number_target_pipelines = len(projects)

    if not os.path.exists(internal_path.training_cache / "dataset-snapshots"):
        os.makedirs(internal_path.training_cache / "dataset-snapshots")

    data = []
    for i in range(0, total_number_target_pipelines):
        data.append((total_number_target_pipelines, projects, i))
    p = Pool(cpu_count())
    values = [item for item in data]
    p.map(_run, values)


if __name__ == "__main__":
    import argparse

    parser = argparse.ArgumentParser()
    parser.add_argument("--tag", type=str, help="Tag for output files and dirs.")
    args = parser.parse_args()
    if args.tag:
        internal_path.training_cache = internal_path.training_cache / args.tag
    main()



# Copyright 2023-2024 The SapientML Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
from collections import OrderedDict, defaultdict

import pandas as pd
from sapientml_core import internal_path
from sapientml_core.seeding.predictor import name_to_label_mapping
from sapientml_core.training.project_corpus import ProjectCorpus
from tqdm import tqdm


class MutationResult:
    """MutationResult class.

    This class loads the mutated results for each pipeline that were already stored in the sapientml_core cache
    and combines all the results in a CSV file and selects the best model.

    """

    def __init__(self, mutation_result_path, project_list):
        self.mutation_result_path = mutation_result_path
        self.project_list = project_list

    def load_results(self):
        """Collects the score for augmented pipelines from exec_info directory.

        Returns
        -------
        results: defaultdict

        """
        results = defaultdict(defaultdict)
        models = list(name_to_label_mapping.keys()) + ["original"]
        execution_root_dir = internal_path.training_cache / "exec_info"

        for i in tqdm(range(0, len(self.project_list))):
            project = self.project_list[i]
            project_exec_dir = execution_root_dir / project.notebook_name
            project_key = project.file_name
            for model in models:
                result_file_path = project_exec_dir / model / "stdout.txt"
                acc, r2 = 0, 0
                if not os.path.exists(result_file_path):
                    results[project_key][model] = 0
                    continue
                with open(result_file_path, "r", encoding="utf-8") as f:
                    lines = f.readlines()
                    for line in lines:
                        for trail in ["Accuracy: ", "R2: "]:
                            data = line
                            if data.count(trail) > 0:
                                data = data[data.index(trail) + len(trail) :].strip()
                                if data.count("%") > 0:
                                    data = data[: data.index("%")]
                                    data = float(data) / 100
                                    if trail == "Accuracy: ":
                                        acc = data
                                    if trail == "R2: ":
                                        r2 = data
                    if project.metric == "accuracy":
                        results[project_key][model] = round(acc, 5)
                    elif project.metric == "r2":
                        results[project_key][model] = round(r2, 5)

            best_models = []
            sorted_results = sorted(results[project_key].items(), key=lambda x: x[1], reverse=True)
            best_value = 0
            for model, value in sorted_results:
                if value > 0 and value >= best_value:
                    best_models.append(model)
                    best_value = value
                else:
                    break

            results[project_key]["best_models"] = best_models

        return results


def main():
    """Fetch the augmented pipeline results and store it in mutation_results.csv."""
    corpus = ProjectCorpus()  # Fetch all project and pipeline details
    mutation_result = MutationResult(internal_path.training_cache, corpus.project_list)
    results = mutation_result.load_results()
    result_list = []
    for key, result in results.items():
        result["file_name"] = key
        result = OrderedDict(result)
        result.move_to_end("file_name", last=False)
        result_list.append(result)
    result_dataframe = pd.DataFrame(result_list)
    result_dataframe.to_csv(internal_path.training_cache / "mutation_results.csv", index=False)


if __name__ == "__main__":
    import argparse

    parser = argparse.ArgumentParser()
    parser.add_argument("--tag", type=str, help="Tag for output files and dirs.")
    args = parser.parse_args()
    if args.tag:
        internal_path.training_cache = internal_path.training_cache / args.tag

    main()



# Copyright 2023-2024 The SapientML Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import json

from sapientml_core import internal_path

LABELS_TO_IGNORE_NOW = {
    "PREPROCESS:DeleteColumns:drop:pandas",
    "PREPROCESS:Category:map:pandas",
    "PREPROCESS:MissingValues:dropna:pandas",
    "PREPROCESS:Category:replace:pandas",
    "PREPROCESS:FeatureSelection:select_dtypes:pandas",
    "PREPROCESS:GenerateColumn:addition:pandas",
}


def main():
    """Removes duplication of labelling orders from dependent_labels.json file.

    This scripts create the dataflow model, i.e., extracts the order of two APIs A and B if there is any.
    There is an order between A --> B if A and B are dependent on each other based on 'dependent_api_extractor.py' and
    A is always followed by B in all piplelines and there is NO case in the corpus where B is followed by A.

    """
    with open(internal_path.training_cache / "dependent_labels.json", "r", encoding="utf-8") as dependent_api_file:
        dependent_labels = json.load(dependent_api_file)

    dependent_order = set()

    for dependent_label_str in dependent_labels.keys():
        dep_str_after_bracket_removal = dependent_label_str.replace("[", "").replace("]", "").replace("'", "")
        parts = dep_str_after_bracket_removal.split(",")
        if (parts[0] in LABELS_TO_IGNORE_NOW) or (parts[1].strip() in LABELS_TO_IGNORE_NOW):
            continue
        first = parts[0].split(":")[1].strip()
        second = parts[1].split(":")[1].strip()
        inverse_order = second + "#" + first
        if first != second:
            if inverse_order in dependent_order:
                dependent_order.remove(inverse_order)
            else:
                dependent_order.add(parts[0].strip() + "#" + parts[1].strip())

    with open(internal_path.training_cache / "label_order.json", "w", encoding="utf-8") as outfile:
        json.dump(list(dependent_order), outfile, indent=4)


if __name__ == "__main__":
    import argparse

    parser = argparse.ArgumentParser()
    parser.add_argument("--tag", type=str, help="Tag for output files and dirs.")
    args = parser.parse_args()
    if args.tag:
        internal_path.training_cache = internal_path.training_cache / args.tag

    main()



from typing import Optional

from sapientml.params import Config, String


class DefaultPreprocessConfig(Config):
    """Configuration arguments for DefaultPreprocess class.

    Attributes
    ----------
    use_pos_list : Optional[list[str]]
      List of parts-of-speech to be used during text analysis.
      This variable is used for japanese texts analysis.
      Select the part of speech below.
      "名詞", "動詞", "形容詞", "形容動詞", "副詞".
    use_word_stemming : bool default True
      Specify whether or not word stemming is used.
      This variable is used for japanese texts analysis.

    """

    use_pos_list: Optional[list[String]] = ["名詞", "動詞", "助動詞", "形容詞", "副詞"]
    use_word_stemming: bool = True



# Copyright 2023-2024 The SapientML Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
from typing import Optional

import numpy as np
from jinja2 import Environment, FileSystemLoader
from sapientml.generator import CodeBlockGenerator
from sapientml.params import Code, Config, Dataset, Task

template_env = Environment(loader=FileSystemLoader(f"{os.path.dirname(__file__)}/templates"), trim_blocks=True)
ROW_THRESHOLD_FOR_SAMPLING = 100000


def _render(tpl, *args, **kwargs):
    code = tpl.render(*args, **kwargs)
    return "\n".join([line for line in code.split("\n") if len(line) > 0]) + "\n\n"


class LocalFileConfig(Config):
    """Configuration arguments for LocalFile class.

    Parameters
    ----------
    id_columns_for_prediction: list[str] or None
        ID columns for prediction.
    """

    id_columns_for_prediction: Optional[list[str]] = None


class LocalFile(CodeBlockGenerator):
    def __init__(self, **kwargs):
        self.config = LocalFileConfig(**kwargs)

    def generate_code(self, dataset: Dataset, task: Task):
        """Generates code including reading files, splitting the dataset, and subsampling.

        This function will update the training_dataframe, validation_dataframe,
        and test_dataframe by droping users' specified ignore_column.

        Parameters
        ----------
        dataset : Dataset
           Object of Dataset class containing the details of dataset.
        task : Task
           Object of Task class containing the details of task.

        Returns
        -------
        dataset: Dataset
            Dataset after dropping users' specified ignore_columns.
        code: Code
            Generated code.
        """

        code = Code()
        code.validation = code.test = code.train = code.predict = "# *** GENERATED PIPELINE ***\n\n"

        dataset, tmp_code = self._generate_code_load(dataset, task)
        code += tmp_code
        dataset, tmp_code = self._generate_code_concat_train_validation(dataset, task)
        code += tmp_code
        dataset, tmp_code = self._generate_code_drop_inf_or_nan(dataset, task)
        code += tmp_code
        dataset, tmp_code = self._generate_code_split(dataset, task)
        code += tmp_code
        dataset, tmp_code = self._generate_code_subsample(dataset, task)
        code += tmp_code
        # set_index msut be before ignore for the case where ignore_columns contain id_columns_for_prediction
        dataset, tmp_code = self._generate_code_set_index(dataset, task)
        code += tmp_code
        dataset, tmp_code = self._generate_code_ignore(dataset, task)
        code += tmp_code
        dataset, tmp_code = self._generate_code_set_validation_as_test(dataset, task)
        code += tmp_code

        return dataset, code

    def _generate_code_load(self, dataset: Dataset, task: Task):
        code = Code()
        tpl = template_env.get_template("load_localfile.py.jinja")
        code.validation += _render(tpl, dataset=dataset, task=task, config=self.config, validation=True)
        code.test += _render(tpl, dataset=dataset, task=task, config=self.config, validation=False)
        tpl = template_env.get_template("load_localfile_train.py.jinja")
        code.train += _render(tpl, dataset=dataset, task=task, config=self.config)
        tpl = template_env.get_template("load_localfile_predict.py.jinja")
        code.predict += _render(tpl, dataset=dataset, task=task, config=self.config)
        return dataset, code

    def _generate_code_set_index(self, dataset: Dataset, task: Task):
        code = Code()
        id_columns_for_prediction = self.config.id_columns_for_prediction
        if id_columns_for_prediction:
            tpl = template_env.get_template("set_index.py.jinja")
            code.test += _render(tpl, id_columns_for_prediction=id_columns_for_prediction)
            code.predict += _render(tpl, id_columns_for_prediction=id_columns_for_prediction)
        return dataset, code

    def _generate_code_ignore(self, dataset: Dataset, task: Task):
        code = Code()
        ignore_columns = dataset.ignore_columns
        if ignore_columns:
            tpl = template_env.get_template("drop_ignore_columns.py.jinja")
            code.validation += _render(tpl, ignore_columns=ignore_columns, train=True, validation=True, test=False)
            code.test += _render(tpl, ignore_columns=ignore_columns, train=True, validation=False, test=True)
            code.train += _render(tpl, ignore_columns=ignore_columns, train=True, validation=False, test=False)
            code.predict += _render(tpl, ignore_columns=ignore_columns, train=False, validation=False, test=True)
            dataset.training_dataframe = dataset.training_dataframe.drop(
                dataset.ignore_columns, axis=1, errors="ignore"
            )
            if dataset.validation_dataframe is not None:
                dataset.validation_dataframe = dataset.validation_dataframe.drop(
                    dataset.ignore_columns, axis=1, errors="ignore"
                )
            if dataset.test_dataframe is not None:
                dataset.test_dataframe = dataset.test_dataframe.drop(dataset.ignore_columns, axis=1, errors="ignore")
        return dataset, code

    def _generate_code_concat_train_validation(self, dataset: Dataset, task: Task):
        code = Code()
        if dataset.validation_dataframe is not None:
            tpl = template_env.get_template("concat_train_validation.py.jinja")
            code.test += _render(tpl)
        return dataset, code

    def _generate_code_split(self, dataset: Dataset, task: Task):
        code = Code()
        tpl = template_env.get_template("split.py.jinja")
        code.validation += _render(tpl, dataset=dataset, task=task, validation=True)
        code.test += _render(tpl, dataset=dataset, task=task, validation=False)
        return dataset, code

    def _generate_code_subsample(self, dataset: Dataset, task: Task):
        code = Code()
        tpl = template_env.get_template("subsample.py.jinja")
        code.validation += _render(tpl, task=task, sample_size=ROW_THRESHOLD_FOR_SAMPLING)
        return dataset, code

    def _generate_code_set_validation_as_test(self, dataset: Dataset, task: Task):
        code = Code()
        tpl = template_env.get_template("set_validation_as_test.py.jinja")
        code.validation += _render(tpl)
        return dataset, code

    def _generate_code_drop_inf_or_nan(self, dataset: Dataset, task: Task):
        code = Code()
        # handle nan or inf value in the target columns.
        drop_flag = dataset.training_dataframe[task.target_columns].isin([np.inf, -np.inf, np.nan]).any(axis=1).any()
        if drop_flag:
            tpl = template_env.get_template("drop_inf_or_nan_rows.py.jinja")
            code.validation += _render(tpl, target_columns=task.target_columns)
            code.test += _render(tpl, target_columns=task.target_columns)
            code.train += _render(tpl, target_columns=task.target_columns)
        return dataset, code



# Copyright 2023-2024 The SapientML Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from sapientml_core import ps_macros
from sapientml_core.design import search_space
from sklearn.tree import DecisionTreeClassifier


def select_k_best_features(X, y):
    """Select the top k explanatory variables.

    Parameters
    ----------
    X : MatrixLike = np.ndarray | pd.DataFrame | spmatrix
        The training input samples.
    y : ArrayLike = numpy.typing.ArrayLike
        The target values

    Returns
    -------
    list
        Returns a list of the top k selected column names.
    """
    from sklearn.feature_selection import SelectKBest, mutual_info_regression

    # Select top 2 features based on mutual info regression
    selector = SelectKBest(mutual_info_regression, k=3)
    selector.fit(X, y)
    return list(X.columns[selector.get_support()])


def select_by_rfe(X, y):
    """Extract the top N(=n_features_to_select) feature values of importance by RFE(Recursive Feature Elimination).

    Parameters
    ----------
    X : MatrixLike = np.ndarray | pd.DataFrame | spmatrix   |
        ArrayLike = numpy.typing.ArrayLike
        The training input samples.
    y : ArrayLike = numpy.typing.ArrayLike
        The target values.

    Returns
    -------
    list
        Returns a list of selected column names.
    """
    from sklearn.feature_selection import RFE

    # #Selecting the Best important features according to Logistic Regression
    rfe_selector = RFE(estimator=DecisionTreeClassifier(), n_features_to_select=2, step=1)
    rfe_selector.fit(X, y)
    return list(X.columns[rfe_selector.get_support()])


def select_from_model(X, y):
    """Select features based on importance weights.

    Parameters
    ----------
    X : MatrixLike = np.ndarray | pd.DataFrame | spmatrix
        The training input samples.
    y : None | ArrayLike = numpy.typing.ArrayLike
        The target values(integers that correspond to classes in classification, real numbers in regression).

    Returns
    -------
    list
        Returns a list of selected column names.
    """
    from sklearn.feature_selection import SelectFromModel

    # #Selecting the Best important features according to Logistic Regression using SelectFromModel
    sfm_selector = SelectFromModel(estimator=DecisionTreeClassifier())
    sfm_selector.fit(X, y)
    return list(X.columns[sfm_selector.get_support()])


def select_sequentially(X, y):
    """Select feature quantity in order and select feature quantity by greedy method.

    Parameters
    ----------
    X : MatrixLike = np.ndarray | pd.DataFrame | spmatrix
        Training vectors
    y : None | ArrayLike = numpy.typing.ArrayLike
        Target values. This parameter may be ignored for unsupervised learning.

    Returns
    -------
    list
        Returns a list of selected column names.
    """
    from sklearn.feature_selection import SequentialFeatureSelector

    # Selecting the Best important features according to Logistic Regression
    sfs_selector = SequentialFeatureSelector(
        estimator=DecisionTreeClassifier(), n_features_to_select=3, cv=10, direction="backward"
    )
    sfs_selector.fit(X, y)
    return list(X.columns[sfs_selector.get_support()])


def select_based_on_correlation(data):
    """Create correlation maps for learning data.

    Parameters
    ----------
    data : dataframe
        Training data

    Returns
    -------
    correlation_map : defaultdict(list)
    """
    from collections import defaultdict

    corr = data.corr(numeric_only=True)
    correlation_map = defaultdict(list)
    for i in range(len(corr.columns)):
        left = corr.columns[i]
        for j in range(i):
            if corr.iloc[i, j] >= 0.25:
                right = corr.columns[j]
                if left[0] != right[0]:
                    correlation_map[left].append(right)

        if len(correlation_map[left]) == 0:
            for j in range(i):
                if corr.iloc[i, j] >= 0.15:
                    right = corr.columns[j]
                    if left[0] != right[0]:
                        correlation_map[left].append(right)

        if len(correlation_map[left]) == 0:
            correlation_map[left] = list(search_space.meta_feature_list)
    return correlation_map


def select_features(label):
    """Return manually selected feature labels.

    Parameters
    ----------
    label : str

    Returns
    -------
    selection_model[label] : list
    """
    selection_model = {
        ps_macros.FILL: [ps_macros.MISSING_PRESENCE],
        # ps_macros.DROP: [ps_macros.MISSING_PRESENCE],
        ps_macros.IN_PLACE_CONVERT: [
            ps_macros.CATG_PRESENCE,
            # ps_macros.IS_TARGET_STR,
            ps_macros.BINARY_CATG_PRESENCE,
            ps_macros.SMALL_CATG_PRESENCE,
            ps_macros.LARGE_CATG_PRESENCE,
        ],
        ps_macros.ONE_HOT: [
            ps_macros.CATG_PRESENCE,
            # ps_macros.IS_TARGET_STR,
            ps_macros.BINARY_CATG_PRESENCE,
            ps_macros.SMALL_CATG_PRESENCE,
            ps_macros.LARGE_CATG_PRESENCE,
        ],
        ps_macros.VECT: [ps_macros.TEXT_PRESENCE],
        ps_macros.MISSING: [ps_macros.MISSING_PRESENCE],
        ps_macros.CATG: [ps_macros.CATG_PRESENCE],
        ps_macros.SCALING: [
            ps_macros.NORMALIZED_MEAN,
            ps_macros.NORMALIZED_STD_DEV,
            ps_macros.NORMALIZED_VARIATION_ACROSS_COLUMNS,
        ],
        ps_macros.DATE: [ps_macros.DATE_PRESENCE],
        ps_macros.LEMMITIZE: [ps_macros.TEXT_PRESENCE],
        ps_macros.BALANCING: [ps_macros.IMBALANCE],
        ps_macros.LOG: [
            ps_macros.MAX_SKEW,
        ],
    }
    return selection_model[label]



# Copyright 2023-2024 The SapientML Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import ast
import copy
import json
import os
import re
import subprocess
import sys
import time
from datetime import datetime
from glob import glob
from pathlib import Path
from shutil import copyfile
from typing import Tuple, Union

from sapientml.executor import PipelineExecutor
from sapientml.generator import CodeBlockGenerator, PipelineGenerator
from sapientml.macros import metric_lower_is_better
from sapientml.params import Code, Dataset, PipelineResult, RunningResult, Task
from sapientml.util.json_util import JSONEncoder
from sapientml.util.logging import setup_logger
from tqdm import tqdm

from . import internal_path
from .adaptation.generation.template_based_adaptation import Adaptation
from .explain.main import process as explain
from .params import SapientMLConfig, SimplePipeline, summarize_dataset
from .seeding.predictor import predict
from .training import project_corpus

if sys.version_info.minor <= 9:
    from importlib_metadata import entry_points
else:
    from importlib.metadata import entry_points

logger = setup_logger()


def add_prefix(filename, prefix):
    """Add prefix to filename if prefix exists.

    Parameters
    ----------
    filename : str
        Filename.
    prefix : str
        Prefix(project_name).

    Returns
    ----------
    str
        If prefix exists, return f"{prefix}_{filename}",
        If prefix does not exist, return only filename.
    """
    if not prefix:
        return filename
    return f"{prefix}_{filename}"


class SapientMLGenerator(PipelineGenerator, CodeBlockGenerator):
    """
    Own the main functions for generating the pipeline.
    """

    def __init__(self, datastore="localfile", preprocess="default", **kwargs):
        self.config = SapientMLConfig(**kwargs)
        self.config.post_init()
        eps_datastore = entry_points(group="sapientml.datastore")
        eps_preprocess = entry_points(group="sapientml.preprocess")
        self.datastore = eps_datastore[datastore].load()(**kwargs)
        self.preprocess = eps_preprocess[preprocess].load()(**kwargs)

    def train(self, tag=None, num_parallelization=200):
        """Run meta-training.

        Parameters
        ----------
        tag : str | None
            The tag of souce code.
            If tag is set, the traning results will be saved into <core/sapientml_core/.cache/tag/>.
            Else if tag is not set, the traning results will be saved into <core/sapientml_core/.cache/>.
        num_parallelization : int | 200
            Number of parallelization.
            Default value is 200.
        """

        def _exec(cmd):
            logger.info(f"Executing {cmd}")
            proc = subprocess.Popen(
                cmd,
                shell=True,
                stdout=subprocess.PIPE,
                stderr=subprocess.PIPE,
                text=True,
            )
            stdout, stderr = proc.communicate()
            logger.debug("STDOUT: {}".format(stdout))
            logger.error("STDERR:{}".format(stderr))
            return proc

        def _wait_for(file_pattern, n_files=1):
            logger.info(f"Waiting for {file_pattern}" + (f" * {n_files}" if n_files > 1 else ""))
            latest = 0
            with tqdm(total=n_files) as pbar:
                pbar.update(latest)
                while latest < n_files:
                    time.sleep(2)
                    k = len(glob(f"{internal_path.training_cache}/{file_pattern}"))
                    if k > latest:
                        pbar.update(k - latest)
                        latest = k

        corpus = project_corpus.ProjectCorpus()
        projects = corpus.project_list

        if tag:
            internal_path.training_cache = internal_path.training_cache / tag

        os.makedirs(internal_path.training_cache, exist_ok=True)

        start = datetime.now()

        logger.info(f"Start: {start}")

        # Step-1
        _exec(
            f"PYTHONPATH=. python {internal_path.sapientml_core_root}/training/denoising/static_analysis_of_columns.py --tag={tag}"
        )
        _wait_for("static_info.json")
        _exec(
            f"PYTHONPATH=. python {internal_path.sapientml_core_root}/training/denoising/dataset_snapshot_extractor.py --tag={tag}"
        )
        _wait_for("dataset-snapshots/*.txt", len(projects))

        # Step-2
        procs = []
        for x in range(num_parallelization):
            proc = _exec(
                f"PYTHONPATH=. python {internal_path.sapientml_core_root}/training/augmentation/mutation_runner.py {num_parallelization} {x} {tag}"
            )
            procs.append(proc)

        _wait_for("exec_info/mutation_*.finished", num_parallelization)

        for p in procs:
            if p.poll() is None:
                logger.info("Kill alive processes")
                p.terminate()
                logger.info("Done, exit status:{}".format(proc.poll()))

        _exec(
            f"PYTHONPATH=. python {internal_path.sapientml_core_root}/training/denoising/determine_used_features.py --tag={tag}"
        )
        _wait_for("feature_analysis_summary.json")
        _exec(
            f"PYTHONPATH=. python {internal_path.sapientml_core_root}/training/augmentation/mutation_results.py --tag={tag}"
        )

        # Step-3
        _exec(f"PYTHONPATH=. python {internal_path.sapientml_core_root}/training/meta_feature_extractor.py --tag={tag}")
        _wait_for("*_metafeatures_training.csv", 2)

        # Step-4
        _exec(f"PYTHONPATH=. python {internal_path.sapientml_core_root}/training/pp_model_trainer.py --tag={tag}")
        _exec(f"PYTHONPATH=. python {internal_path.sapientml_core_root}/training/meta_model_trainer.py --tag={tag}")

        _wait_for("*.pkl", 3)

        # Step-5
        _exec(
            f"PYTHONPATH=. python {internal_path.sapientml_core_root}/training/dataflowmodel/dependent_api_extractor.py --tag={tag}"
        )
        _wait_for("dependent_labels.json")
        _exec(
            f"PYTHONPATH=. python {internal_path.sapientml_core_root}/training/dataflowmodel/determine_label_order.py --tag={tag}"
        )
        _wait_for("label_order.json")

        end = datetime.now()
        logger.info(f"End: {end}")

        diff_time = end - start
        logger.info(f"Training time cost: {diff_time}")

    def generate_pipeline(self, dataset: Dataset, task: Task):
        """Generate candidate scripts, execute them and select the best one.

        Parameters
        ----------
        dataset : Dataset
            Dataset class.
        task : Task
            Task class.

        Returns
        ----------
        (best_pipeline, best_pipeline_score) : Tuple(list | None, PipelineResult)
            best_pipeline: Best pipeline.
            best_pipline_score: The score of the best pipeline.
        candidate_scripts : list
            Candidate scripts with scores.
        """
        self.dataset = dataset
        self.task = task

        logger.info("Generating pipelines...")
        dataset, datastore_block = self.datastore.generate_code(dataset, task)
        dataset.check_dataframes(task.target_columns)
        dataset, preprocess_block = self.preprocess.generate_code(dataset, task)
        code_block = datastore_block + preprocess_block
        dataset, sapientml_results = self.generate_code(dataset, task)

        result_pipelines: list[Code] = []
        for pipeline in sapientml_results:
            pipeline.validation = code_block.validation + pipeline.validation
            pipeline.test = code_block.test + pipeline.test
            pipeline.train = code_block.train + pipeline.train
            pipeline.predict = code_block.predict + pipeline.predict
            result_pipelines.append(pipeline)

        logger.info("Executing generated pipelines...")
        executor = PipelineExecutor()
        self.execution_results = executor.execute(
            result_pipelines,
            self.config.initial_timeout,
            Path(dataset.output_dir),
            self.config.cancel,
        )

        logger.info("Evaluating execution results of generated pipelines...")
        lower_is_better = self.task.adaptation_metric in metric_lower_is_better
        self.evaluate(self.execution_results, lower_is_better)

        return (self._best_pipeline, self._best_pipeline_score), self._candidate_scripts

    def generate_code(self, dataset: Dataset, task: Task) -> Tuple[Dataset, list[SimplePipeline]]:
        """
        Generate code including preprocessing and machine learning models using meta-features.

        Parameters
        ----------
        dataset : Dataset
            Dataset class.
        task : Task
            Task class.

        Returns
        ----------
        dataset : Dataset
            Dataset class.
        pipelines : list[SimplePipeline]
            Generated code. Different pipelines contain different models.
        """
        df = dataset.training_dataframe
        # Generate the meta-features
        logger.info("Generating meta features...")
        dataset_summary = summarize_dataset(df, task)  # type: ignore

        labels = predict(task, dataset_summary)
        adapt = Adaptation(
            labels,
            task,
            dataset_summary,
            self.config,
        )
        pipelines = adapt.run_adaptation()

        return dataset, pipelines

    def evaluate(self, execution_results: list[tuple[Code, RunningResult]], lower_is_better: bool = False) -> None:
        """Evaluate execution results of generated pipelines.

        Parameters
        ----------
        execution_results : list[tuple[Code, RunningResult]]
            Execution results containing executed code and resulted output texts.
        lower_is_better : bool, default False
            Specify reverse=(not lower_is_better) for the argument of the sorted method.

        Returns
        ----------
        None

        """
        self._best_pipeline = None
        self._best_pipeline_score = PipelineResult(score=None, metric=None, best_params=None)
        candidate_scripts = []
        for pipeline, result in execution_results:
            if result.returncode == 0:
                pipeline_score = self._parse_pipeline_output(result.output)
            else:
                pipeline_score = PipelineResult(score=None, metric=None, best_params=None)
            candidate_scripts.append((pipeline, pipeline_score))
        self._candidate_scripts = candidate_scripts

        # When an error occurs while running a pipeline, the score becomes None
        error_pipelines = [pipeline for pipeline in candidate_scripts if pipeline[1].score is None]

        # If none of them have the score, stop ranking them
        if len(candidate_scripts) == len(error_pipelines):
            return

        # sort descending
        succeeded_scripts = sorted(
            [x for x in candidate_scripts if x[1].score is not None],
            key=lambda x: x[1].score,
            reverse=(not lower_is_better),
        )
        failed_scripts = [x for x in candidate_scripts if x[1].score is None]

        ranked_candidate_scripts = succeeded_scripts + failed_scripts
        best_pipeline_tuple = ranked_candidate_scripts[0]
        if best_pipeline_tuple is None:
            return

        best_pipeline = copy.deepcopy(best_pipeline_tuple[0])
        if best_pipeline_tuple[1].best_params is not None:
            best_pipeline.test = best_pipeline.test.replace(
                "best_params = study.best_params", "best_params = " + str(best_pipeline_tuple[1].best_params)
            )
            best_pipeline.train = best_pipeline.train.replace(
                "best_params = study.best_params", "best_params = " + str(best_pipeline_tuple[1].best_params)
            )
        self._best_pipeline = best_pipeline
        self._best_pipeline_score = best_pipeline_tuple[1]

    @staticmethod
    def _parse_pipeline_output(output: str):
        score = None
        best_params = None
        metric = None
        output_lines = output.splitlines()
        try:
            for line in output_lines:
                if re.match("best params: ", line):
                    best_params = ast.literal_eval(re.findall("best params: (.+)", line)[0])
                elif re.match("RESULT: ", line):
                    parts = [x.strip() for x in line.split(":")]
                    metric = parts[-2].strip().split(" ")[0]
                    score = float(parts[-1])
        except Exception:
            pass
        return PipelineResult(score=score, metric=metric, best_params=best_params)

    def save(self, output_dir: Union[Path, str]):
        """Save the script to output_dir.

        Parameters
        ----------
        output_dir : Union[Path, str]
            Output directory path.
        """
        self.output_dir = Path(output_dir)
        self.output_dir.mkdir(parents=True, exist_ok=True)

        candidate_scripts = self._candidate_scripts
        if candidate_scripts:
            if self._best_pipeline:
                script_body = self._best_pipeline.test
                with open(
                    self.output_dir / add_prefix("final_script.py", self.config.project_name), "w", encoding="utf-8"
                ) as f:
                    f.write(script_body)

                script_body = self._best_pipeline.train
                with open(
                    self.output_dir / add_prefix("final_train.py", self.config.project_name), "w", encoding="utf-8"
                ) as f:
                    f.write(script_body)

                script_body = self._best_pipeline.predict
                with open(
                    self.output_dir / add_prefix("final_predict.py", self.config.project_name), "w", encoding="utf-8"
                ) as f:
                    f.write(script_body)

                with open(
                    self.output_dir / (add_prefix("final_script", self.config.project_name) + ".out.json"),
                    "w",
                    encoding="utf-8",
                ) as f:
                    json.dump(self._best_pipeline_score.__dict__, f, cls=JSONEncoder, indent=4)
            else:
                logger.warning("All candidate scripts failed. Final script is not saved.")
                raise RuntimeError("All candidate scripts failed. Final script is not saved.")

            # copy libs
            lib_path = self.output_dir / "lib"
            lib_path.mkdir(exist_ok=True)

            eps = entry_points(group="sapientml.export_modules")
            for ep in eps:
                if ep.name in [self.__class__.__name__, "sample-dataset"]:
                    for file in glob(f"{ep.load().__path__[0]}/*.py"):
                        copyfile(file, lib_path / Path(file).name)

            for index, (script, detail) in enumerate(candidate_scripts, start=1):
                script_body = script.validation
                with open(self.output_dir / f"{index}_script.py", "w", encoding="utf-8") as f:
                    f.write(script_body)

        self.debug_info = {}
        for i, candidate in enumerate(candidate_scripts, start=1):
            info = {"content": candidate[0].model_dump(), "run_info": candidate[1].__dict__}
            self.debug_info[i] = info

        if self.config.debug:
            with open(
                self.output_dir / add_prefix("run_info.json", self.config.project_name), "w", encoding="utf-8"
            ) as f:
                json.dump(self.debug_info, f, cls=JSONEncoder, indent=4)

        if self.config.add_explanation:
            self.add_explanation()

    def add_explanation(self):
        """Call the explain function."""
        explain(
            visualization=True,
            eda=True,
            dataframe=self.dataset.training_dataframe,
            script_path=(self.output_dir / add_prefix("final_script.py", self.config.project_name))
            .absolute()
            .as_posix(),
            target_columns=self.task.target_columns,
            problem_type=self.task.task_type,
            ignore_columns=self.dataset.ignore_columns,
            skeleton=self._best_pipeline.labels,
            explanation=self._best_pipeline.pipeline_json,
            run_info=self.debug_info,
            internal_execution=True,
            timeout=self.config.timeout_for_test,
            cancel=self.config.cancel,
        )



# Copyright 2023-2024 The SapientML Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


name_to_label_mapping = {
    "random forest": {
        "c": "MODEL:Classifier:RandomForestClassifier:sklearn",
        "r": "MODEL:Regressor:RandomForestRegressor:sklearn",
    },
    "extra tree": {
        "c": "MODEL:Classifier:ExtraTreesClassifier:sklearn",
        "r": "MODEL:Regressor:ExtraTreesRegressor:sklearn",
    },
    "lightgbm": {"c": "MODEL:Classifier:LGBMClassifier:lightgbm", "r": "MODEL:Regressor:LGBMRegressor:lightgbm"},
    "xgboost": {"c": "MODEL:Classifier:XGBClassifier:xgboost", "r": "MODEL:Regressor:XGBRegressor:xgboost"},
    "catboost": {
        "c": "MODEL:Classifier:CatBoostClassifier:catboost",
        "r": "MODEL:Regressor:CatBoostRegressor:catboost",
    },
    "gradient boosting": {
        "c": "MODEL:Classifier:GradientBoostingClassifier:sklearn",
        "r": "MODEL:Regressor:GradientBoostingRegressor:sklearn",
    },
    "adaboost": {"c": "MODEL:Classifier:AdaBoostClassifier:sklearn", "r": "MODEL:Regressor:AdaBoostRegressor:sklearn"},
    "decision tree": {
        "c": "MODEL:Classifier:DecisionTreeClassifier:sklearn",
        "r": "MODEL:Regressor:DecisionTreeRegressor:sklearn",
    },
    "svm": {"c": "MODEL:Classifier:SVC:sklearn", "r": "MODEL:Regressor:SVR:sklearn"},
    "linear svm": {"c": "MODEL:Classifier:LinearSVC:sklearn", "r": "MODEL:Regressor:LinearSVR:sklearn"},
    "logistic/linear regression": {
        "c": "MODEL:Classifier:LogisticRegression:sklearn",
        "r": "MODEL:Regressor:LinearRegression:sklearn",
    },
    "lasso": {"r": "MODEL:Regressor:Lasso:sklearn"},
    "sgd": {"c": "MODEL:Classifier:SGDClassifier:sklearn", "r": "MODEL:Regressor:SGDRegressor:sklearn"},
    "mlp": {"c": "MODEL:Classifier:MLPClassifier:sklearn", "r": "MODEL:Regressor:MLPRegressor:sklearn"},
    "multinomial nb": {"c": "MODEL:Classifier:MultinomialNB:sklearn"},
    "gaussian nb": {"c": "MODEL:Classifier:GaussianNB:sklearn"},
    "bernoulli nb": {"c": "MODEL:Classifier:BernoulliNB:sklearn"},
}


def map_label_to_name():
    """
    Assign several internal labels to each ML component.

    Returns
    ----------
    label_to_name : dict[str, str]
        Assigned result.
    """
    label_to_name = {"MODEL:Classifier:LGBMClassifier:lgbm": "lightgbm", "MODEL:Regressor:train:xgboost": "xgboost"}
    for k, v in name_to_label_mapping.items():
        for k1, v1 in v.items():
            label_to_name[v1] = k
    return label_to_name



# Copyright 2023-2024 The SapientML Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import ast
import os

from sapientml.util.logging import setup_logger

logger = setup_logger()

"""
    I edited khsamaha_aviation-accident-database-synopses/predict-percentage-of-fatal-injuries.py
    since the original one is not complete and cannot be parsed
"""


model_match_keywords = [
    ("RandomForestRegressor", "random forest", "regression"),
    ("RandomForestClassifier", "random forest", "classification"),
    ("BaggingClassifier", "bagging", "classification"),
    ("BaggingRegressor", "bagging", "regression"),
    ("AdaBoostRegressor", "adaboost", "regression"),
    ("LinearRegression", "logistic/linear regression", "regression"),
    ("LogisticRegression", "logistic/linear regression", "classification"),
    ("SVC", "svm", "classification"),
    ("SVR", "svm", "regression"),
    ("MultinomialNB", "multinomial nb", "classification"),
    ("GaussianNB", "gaussian nb", "classification"),
    ("XGBClassifier", "xgboost", "classification"),
    ("XGBRegressor", "xgboost", "regression"),
    ("xgboost", "xgboost", "classification"),
    ("DecisionTreeClassifier", "decision tree", "classification"),
    ("DecisionTreeRegressor", "decision tree", "regression"),
    ("Ridge", "ridge", "regression"),
    ("KNeighborsClassifier", "knn", "classification"),
    ("Lasso", "lasso", "regression"),
    ("LinearSVC", "linear svm", "classification"),
    ("SVC(kernel='linear')", "linear svm", "classification"),
    # view Gaussian family GLM as ridge regressor
    ("GLM", "ridge", "regression"),
    ("OLS", "logistic/linear regression", "regression"),
    ("GradientBoostingClassifier", "gradient boosting", "classification"),
    ("GradientBoostingRegressor", "gradient boosting", "regression"),
    ("ExtraTreesRegressor", "extra tree", "regression"),
    ("LGBMClassifier", "lightgbm", "classification"),
    ("Logit", "logistic/linear regression", "classification"),
    ("MLPClassifier", "mlp", "classification"),
    # ad-hoc
    ("RFE", "logistic/linear regression", "classification"),
    # new
    ("KMeans", "kmeans", "classification"),
    ("ElasticNet", "logistic/linear regression", "regression"),
    ("MLPRegressor", "mlp", "regression"),
    ("AdaBoostClassifier", "adaboost", "classification"),
    ("SGDRegressor", "sgd", "regression"),
    ("SGDClassifier", "sgd", "classification"),
    ("ExtraTreesClassifier", "extra tree", "classification"),
    ("Perceptron", "sgd", "classification"),
    ("LGBMRegressor", "lightgbm", "regression"),
    ("KNeighborsRegressor", "knn", "regression"),
]

# only GridSearchCV is hard to detect model type
# only RandomizedSearchCV is hard to mutate
# only VotingClassifier is hard to mutate
# only BaggingClassifier is hard to mutate

# TODO:
#  need to handle xgboost's DMatrix (minor: only affect one dataset)


order = [
    "random forest",
    "extra tree",
    "lightgbm",
    "xgboost",
    "catboost",
    "gradient boosting",
    "adaboost",
    "decision tree",
    "svm",
    "linear svm",
    "logistic/linear regression",
    "lasso",
    "sgd",
    "ridge",
    "knn",
    "bagging",  # new
    "mlp",
    "multinomial nb",
    "gaussian nb",
    "bernoulli nb",
]

# not alter the prediction statement now
templates = {
    # 'model type': (definition statement, training statement)
    ("random forest", "c"): (
        """from sklearn.ensemble import RandomForestClassifier
#1 = RandomForestClassifier()""",
        "#2 = #1.fit(#XT, #YT)",
    ),
    ("random forest", "r"): (
        """from sklearn.ensemble import RandomForestRegressor
from sklearn.multioutput import MultiOutputRegressor
import pandas
if isinstance(#YT, pandas.DataFrame) and len(#YT.columns) > 1:
    #1 = MultiOutputRegressor(RandomForestRegressor())
else:
    #1 = RandomForestRegressor()""",
        "#2 = #1.fit(#XT, #YT)",
    ),
    ("decision tree", "c"): (
        """from sklearn.tree import DecisionTreeClassifier
#1 = DecisionTreeClassifier(random_state=0)""",
        "#2 = #1.fit(#XT, #YT)",
    ),
    ("decision tree", "r"): (
        """from sklearn.tree import DecisionTreeRegressor
from sklearn.multioutput import MultiOutputRegressor
import pandas
if isinstance(#YT, pandas.DataFrame) and len(#YT.columns) > 1:
    #1 = MultiOutputRegressor(DecisionTreeRegressor(random_state=0))
else:
    #1 = DecisionTreeRegressor(random_state=0)""",
        "#2 = #1.fit(#XT, #YT)",
    ),
    ("extra tree", "c"): (
        """from sklearn.ensemble import ExtraTreesClassifier
#1 = ExtraTreesClassifier(random_state=0)""",
        "#2 = #1.fit(#XT, #YT)",
    ),
    ("extra tree", "r"): (
        """from sklearn.ensemble import ExtraTreesRegressor
from sklearn.multioutput import MultiOutputRegressor
import pandas
if isinstance(#YT, pandas.DataFrame) and len(#YT.columns) > 1:
    #1 = MultiOutputRegressor(ExtraTreesRegressor(random_state=0))
else:
    #1 = ExtraTreesRegressor(random_state=0)""",
        "#2 = #1.fit(#XT, #YT)",
    ),
    # =================================
    ("xgboost", "c"): (
        """import xgboost as xgb
#1 = xgb.XGBClassifier()""",
        "#2 = #1.fit(#XT, #YT)",
    ),
    ("xgboost", "r"): (
        """import xgboost as xgb
from sklearn.multioutput import MultiOutputRegressor
import pandas
if isinstance(#YT, pandas.DataFrame) and len(#YT.columns) > 1:
    #1 = MultiOutputRegressor(xgb.XGBRegressor())
else:
    #1 = xgb.XGBRegressor()""",
        "#2 = #1.fit(#XT, #YT)",
    ),
    ("catboost", "c"): (
        """import catboost as cat
#1 = cat.CatBoostClassifier()""",
        "#2 = #1.fit(#XT, #YT)",
    ),
    ("catboost", "r"): (
        """import catboost as cat
from sklearn.multioutput import MultiOutputRegressor
import pandas
if isinstance(#YT, pandas.DataFrame) and len(#YT.columns) > 1:
    #1 = MultiOutputRegressor(cat.CatBoostRegressor())
else:
    #1 = cat.CatBoostRegressor()""",
        "#2 = #1.fit(#XT, #YT)",
    ),
    ("gradient boosting", "c"): (
        """from sklearn.ensemble import GradientBoostingClassifier
#1 = GradientBoostingClassifier()""",
        "#2 = #1.fit(#XT, #YT)",
    ),
    ("gradient boosting", "r"): (
        """from sklearn.ensemble import GradientBoostingRegressor
from sklearn.multioutput import MultiOutputRegressor
import pandas
if isinstance(#YT, pandas.DataFrame) and len(#YT.columns) > 1:
    #1 = MultiOutputRegressor(GradientBoostingRegressor())
else:
    #1 = GradientBoostingRegressor()""",
        "#2 = #1.fit(#XT, #YT)",
    ),
    ("adaboost", "c"): (
        """from sklearn.ensemble import AdaBoostClassifier
#1 = AdaBoostClassifier()""",
        "#2 = #1.fit(#XT, #YT)",
    ),
    ("adaboost", "r"): (
        """from sklearn.ensemble import AdaBoostRegressor
from sklearn.multioutput import MultiOutputRegressor
import pandas
if isinstance(#YT, pandas.DataFrame) and len(#YT.columns) > 1:
    #1 = MultiOutputRegressor(AdaBoostRegressor())
else:
    #1 = AdaBoostRegressor()""",
        "#2 = #1.fit(#XT, #YT)",
    ),
    ("lightgbm", "c"): (
        """from lightgbm import LGBMClassifier
#1 = LGBMClassifier()""",
        """import pandas
if isinstance(#XT, pandas.DataFrame): #XT.columns = [str(i) for i in range(len(#XT.columns))]
#2 = #1.fit(#XT, #YT)""",
        """import pandas
if isinstance(#XS, pandas.DataFrame): #XS.columns = [str(i) for i in range(len(#XS.columns))]""",
    ),
    ("lightgbm", "r"): (
        """from lightgbm import LGBMRegressor
import pandas
from sklearn.multioutput import MultiOutputRegressor
if isinstance(#YT, pandas.DataFrame) and len(#YT.columns) > 1:
    #1 = MultiOutputRegressor(LGBMRegressor())
else:
    #1 = LGBMRegressor()""",
        """import pandas
if isinstance(#XT, pandas.DataFrame):
    #XT.columns = [str(i) for i in range(len(#XT.columns))]
#2 = #1.fit(#XT, #YT)""",
        """import pandas
if isinstance(#XS, pandas.DataFrame):
    #XS.columns = [str(i) for i in range(len(#XS.columns))]""",
    ),
    # =================================
    ("svm", "c"): (
        """from sklearn.svm import SVC
#1 = SVC()""",
        "#2 = #1.fit(#XT, #YT)",
    ),
    ("svm", "r"): (
        """from sklearn.svm import SVR
from sklearn.multioutput import MultiOutputRegressor
import pandas
if isinstance(#YT, pandas.DataFrame) and len(#YT.columns) > 1:
    #1 = MultiOutputRegressor(SVR())
else:
    #1 = SVR()""",
        "#2 = #1.fit(#XT, #YT)",
    ),
    ("linear svm", "c"): (
        """from sklearn.svm import LinearSVC
#1 = LinearSVC()""",
        "#2 = #1.fit(#XT, #YT)",
    ),
    ("linear svm", "r"): (
        """from sklearn.svm import LinearSVR
from sklearn.multioutput import MultiOutputRegressor
import pandas
if isinstance(#YT, pandas.DataFrame) and len(#YT.columns) > 1:
    #1 = MultiOutputRegressor(LinearSVR())
else:
    #1 = LinearSVR()""",
        "#2 = #1.fit(#XT, #YT)",
    ),
    ("logistic/linear regression", "c"): (
        """from sklearn.linear_model import LogisticRegression
#1 = LogisticRegression()""",
        "#2 = #1.fit(#XT, #YT)",
    ),
    ("logistic/linear regression", "r"): (
        """from sklearn.linear_model import LinearRegression
from sklearn.multioutput import MultiOutputRegressor
import pandas
if isinstance(#YT, pandas.DataFrame) and len(#YT.columns) > 1:
    #1 = MultiOutputRegressor(LinearRegression())
else:
    #1 = LinearRegression()""",
        "#2 = #1.fit(#XT, #YT)",
    ),
    ("lasso", "r"): (
        """from sklearn.linear_model import Lasso
from sklearn.multioutput import MultiOutputRegressor
import pandas
if isinstance(#YT, pandas.DataFrame) and len(#YT.columns) > 1:
    #1 = MultiOutputRegressor(Lasso())
else:
    #1 = Lasso()""",
        "#2 = #1.fit(#XT, #YT)",
    ),
    ("sgd", "c"): (
        """from sklearn.linear_model import SGDClassifier
#1 = SGDClassifier()""",
        "#2 = #1.fit(#XT, #YT)",
    ),
    ("sgd", "r"): (
        """from sklearn.linear_model import SGDRegressor
from sklearn.multioutput import MultiOutputRegressor
import pandas
if isinstance(#YT, pandas.DataFrame) and len(#YT.columns) > 1:
    #1 = MultiOutputRegressor(SGDRegressor())
else:
    #1 = SGDRegressor()""",
        "#2 = #1.fit(#XT, #YT)",
    ),
    # =================================
    ("mlp", "c"): (
        """from sklearn.neural_network import MLPClassifier
#1 = MLPClassifier()""",
        "#2 = #1.fit(#XT, #YT)",
    ),
    ("mlp", "r"): (
        """from sklearn.neural_network import MLPRegressor
from sklearn.multioutput import MultiOutputRegressor
import pandas
if isinstance(#YT, pandas.DataFrame) and len(#YT.columns) > 1:
    #1 = MultiOutputRegressor(MLPRegressor())
else:
    #1 = MLPRegressor()""",
        "#2 = #1.fit(#XT, #YT)",
    ),
    # =================================
    ("multinomial nb", "c"): (
        """from sklearn.naive_bayes import MultinomialNB
#1 = MultinomialNB()""",
        "#2 = #1.fit(#XT, #YT)",
    ),
    ("gaussian nb", "c"): (
        """from sklearn.naive_bayes import GaussianNB
#1 = GaussianNB()""",
        "#2 = #1.fit(#XT, #YT)",
    ),
    ("bernoulli nb", "c"): (
        """from sklearn.naive_bayes import BernoulliNB
#1 = BernoulliNB()""",
        "#2 = #1.fit(#XT, #YT)",
    ),
    # ===================================
    ("ridge", "c"): (
        """from sklearn.linear_model import RidgeClassifier
#1 = RidgeClassifier()""",
        "#2 = #1.fit(#XT, #YT)",
    ),
    ("ridge", "r"): (
        """from sklearn.linear_model import Ridge
from sklearn.multioutput import MultiOutputRegressor
import pandas
if isinstance(#YT, pandas.DataFrame) and len(#YT.columns) > 1:
    #1 = MultiOutputRegressor(Ridge())
else:
    #1 = Ridge()""",
        "#2 = #1.fit(#XT, #YT)",
    ),
    ("knn", "c"): (
        """from sklearn.neighbors import KNeighborsClassifier
#1 = KNeighborsClassifier()""",
        "#2 = #1.fit(#XT, #YT)",
    ),
    ("knn", "r"): (
        """from sklearn.neighbors import KNeighborsRegressor
from sklearn.multioutput import MultiOutputRegressor
import pandas
if isinstance(#YT, pandas.DataFrame) and len(#YT.columns) > 1:
    #1 = MultiOutputRegressor(KNeighborsRegressor())
else:
    #1 = KNeighborsRegressor()""",
        "#2 = #1.fit(#XT, #YT)",
    ),
    ("bagging", "c"): (
        """from sklearn.ensemble import BaggingClassifier
#1 = BaggingClassifier()""",
        "#2 = #1.fit(#XT, #YT)",
    ),
    ("bagging", "r"): (
        """from sklearn.ensemble import BaggingRegressor
from sklearn.multioutput import MultiOutputRegressor
import pandas
if isinstance(#YT, pandas.DataFrame) and len(#YT.columns) > 1:
    #1 = MultiOutputRegressor(BaggingRegressor())
else:
    #1 = BaggingRegressor()""",
        "#2 = #1.fit(#XT, #YT)",
    ),
}


def _catch_target_name(assign_stmt: ast.Assign):
    # only digest the first node
    target = assign_stmt.targets[0]

    if isinstance(target, ast.Name):
        return target.id
    else:
        for node_name in ast.walk(target):
            if isinstance(node_name, ast.Name):
                return node_name.id


def _catch_func_name(call_stmt):
    if not isinstance(call_stmt, ast.Call):
        return None
    else:
        func = call_stmt.func
        if isinstance(func, ast.Name):
            func = func.id
        elif isinstance(func, ast.Attribute):
            func = func.attr
        elif isinstance(func, ast.Lambda):
            func = None
        return func


def _catch_method_varname(node):
    if isinstance(node, ast.Name):
        return node.id, list()
    elif isinstance(node, ast.Attribute) and isinstance(node.value, ast.Name):
        return ast.unparse(node), list()
    for nownode in ast.walk(node):
        if isinstance(nownode, ast.Call) and ast.unparse(nownode.func) not in [
            "print"
        ]:  # filter out useless function call
            method_name = ast.unparse(nownode.func)
            variable_names = [ast.unparse(arg_node) for arg_node in nownode.args] + [
                ast.unparse(kw_node.value) for kw_node in nownode.keywords
            ]
            return method_name, variable_names
    return None, list()


class MutationException(Exception):
    """MutationException class"""

    pass


# init: check for what model types, the templates are missing
for o in order:
    for t in ["c", "r"]:
        if (o, t) not in templates:
            logger.info(f"template for {(o, t)} is missing")


# Parse different entities from pipeline using AST library.
def pipeline_analysis(in_path):
    """Parse different entities from pipeline using AST library.

    Parameters
    ----------
    in_path: PosixPath
        It contains path from pathlib.PosixPath

    Returns
    -------
        m_type : str
            It returns a string variable.
        task_type : str
            It returns a string variable.
        def_stmt : tuple
            It returns a tuple.
        train_stmt : tuple
            It returns a tuple.
        test_stmt : tuple
            It returns a tuple.
        model_pred_var : str
            It returns a string variable.
        model_train_var : str
            It returns a string variable.
        train_X.strip() : str
            It returns a string variable.
        train_Y.strip() : str
            It returns a string variable.
        test_X.strip() :  str
            It returns a string variable.
    """
    with open(in_path, "r", encoding="utf-8") as f:
        original_script = f.read()

    tree = ast.parse(source=original_script, filename=in_path)  # parse pipeline to tree

    possible_training_stmts = list()
    possible_test_stmts = list()

    for node in ast.walk(tree):
        if isinstance(node, ast.Call):
            name = None
            func = node.func
            if isinstance(func, ast.Call):
                func = func.func
            if isinstance(func, ast.Name):
                func = func.id
            elif isinstance(func, ast.Attribute):
                if isinstance(func.value, ast.Name):
                    name = func.value.id
                elif isinstance(func.value, ast.AST):
                    for node_name in ast.walk(func.value):
                        if isinstance(node_name, ast.Name):
                            name = node_name.id
                else:
                    raise Exception("Should not happen...")

                func = func.attr
            elif isinstance(func, ast.Lambda):
                continue
            else:
                raise Exception("Unknown func in call stmt", ast.unparse(func), ast.dump(func))
            # support xgb.train is non-trivial, and has benefit only on one notebook: hubh0799_churn-modelling/churn-modelling-xgboost-aucroc-0-902347.py
            if func in ["fit"]:
                possible_training_stmts.append((name, func, node.lineno))
            if func in ["predict", "predict_proba"]:
                possible_test_stmts.append((name, func, node.lineno))

    if len(possible_training_stmts) == 0 or len(possible_test_stmts) == 0:
        raise MutationException("No training or test statement found")
    else:
        pass

    # choose the last possible testing statement as the testing statement
    test_stmt = sorted(possible_test_stmts, key=lambda x: x[2], reverse=True)[0]
    train_stmt = None
    def_stmt = None
    model_pred_var = model_train_var = test_stmt[0]

    # determine real training statement and model definition statement
    possible_training_line_no = [x[2] for x in possible_training_stmts]
    import_name_mapping = dict()
    def_stmts = dict()
    for node in tree.body:
        if isinstance(node, ast.Import) or isinstance(node, ast.ImportFrom):
            names = node.names
            for a in names:
                if a.asname is not None:
                    import_name_mapping[a.asname] = a.name
        if isinstance(node, ast.Assign):
            if node.lineno in possible_training_line_no:
                now_assigned_var = _catch_target_name(node)
                # if the assigned one matches the predict var name, then proceed
                if now_assigned_var == test_stmt[0]:
                    now_train_var = [x[0] for x in possible_training_stmts if x[2] == node.lineno][0]
                    if now_train_var not in def_stmts:
                        # means the training statement uses a in-place defined model
                        # in this case, we view the training var name as the same as the predict var name
                        train_stmt = [x for x in possible_training_stmts if x[2] == node.lineno][0]
                        train_stmt = (model_train_var, train_stmt[1], train_stmt[2])
                        def_stmt = (model_train_var, train_stmt[1], train_stmt[2])
                    else:
                        model_train_var = now_train_var
                        train_stmt = [x for x in possible_training_stmts if x[2] == node.lineno][0]
                        def_stmt = def_stmts[now_train_var]

            now_target = _catch_target_name(node)
            # if now_target not in def_stmts:
            def_stmts[now_target] = (now_target, _catch_func_name(node.value), node.lineno)

        elif isinstance(node, ast.Expr) and isinstance(node.value, ast.Call):
            if node.lineno in possible_training_line_no:
                now_train_stmt = [x for x in possible_training_stmts if x[2] == node.lineno][0]
                if now_train_stmt[0] == model_train_var:
                    train_stmt = now_train_stmt
                    def_stmt = def_stmts[model_train_var]

    if def_stmt is None or train_stmt is None or test_stmt is None:
        raise MutationException("Unable to locate def stmt or train stmt or test stmt")

    # relocate the test stmt as the closest immediate one after training stmt
    test_stmt = sorted([t for t in possible_test_stmts if t[2] >= train_stmt[2]], key=lambda x: x[2])[0]

    # extract method name and variables in definition statement and training statement
    for node in tree.body:
        if node.lineno == def_stmt[2]:
            dnode = node
            if def_stmt[2] == train_stmt[2]:
                # an inplace definition statement
                # in this case, we need to trim the root calling part
                if isinstance(dnode, ast.Assign) and isinstance(dnode.value, ast.Call):
                    dnode = dnode.value.func
            def_method_name, def_var_names = _catch_method_varname(dnode)
            if def_method_name is None:
                raise MutationException(f"Unusual model definition statement: {ast.unparse(node)}")

        if node.lineno == train_stmt[2]:
            train_method_name, train_var_names = _catch_method_varname(node)

    for call_node in ast.walk(tree):
        if (
            isinstance(call_node, ast.Call)
            and isinstance(call_node.func, ast.Attribute)
            and call_node.func.attr in ["predict", "predict_proba"]
        ):
            test_method_name, test_var_names = _catch_method_varname(call_node)

    # determine model type from definition statement
    model_signature = (
        (def_method_name + " " + " ".join(def_var_names))
        .replace(".", " ")
        .replace("(", " ")
        .replace(")", " ")
        .replace(",", " ")
        .split(" ")
    )
    tmp = list()
    for item in model_signature:
        for k, w in import_name_mapping.items():
            if k == item:
                item = w
        tmp.append(item)
    model_signature = " ".join(tmp)

    kws, m_types, task_types = list(), list(), list()
    for kw, m_type, task_type in model_match_keywords:
        if model_signature.count(kw) > 0:
            kws.append(kw)
            m_types.append(m_type)
            task_types.append(task_type)
    if len(kws) > 1 and "xgboost" in kws:
        # "xgboost" has lower priority than other keywords
        index_of = kws.index("xgboost")
        del kws[index_of]
        del m_types[index_of]
        del task_types[index_of]
    if len(kws) == 2 and "SVC" in kws and ("LinearSVC" in kws or "SVC(kernel='linear')" in kws):
        kws = ["LinearSVC"]
        m_types = ["linear svm"]
        task_types = ["classification"]
    if len(kws) > 1 and "adaboost" in m_types:
        # adaboost has higher priority than other keywords
        index_of = m_types.index("adaboost")
        kws = [kws[index_of]]
        m_types = [m_types[index_of]]
        task_types = [task_types[index_of]]

    if (
        model_signature.count("GridSearch") > 0
        or model_signature.count("GridSearchCV") > 0
        or model_signature.count("VotingClassifier") > 0
        or model_signature.count("VotingRegressor") > 0
        or model_signature.count("RandomizedSearchCV") > 0
        or model_signature.count("Pipeline") > 0
        or model_signature.count("make_pipeline")
    ):
        if len(kws) != 1:
            raise MutationException(
                "For gridsearch, cross validation, pipeline, and voting classifier, without more information we cannot determine the model type"
            )
    else:
        if len(kws) == 0:
            raise MutationException(f"Unsupport model type: {def_method_name}")
        else:
            assert len(kws) == 1

    # model type
    kw = kws[0]
    m_type = m_types[0]
    task_type = task_types[0]
    if kw == "GLM" or kw == "OLS" or kw == "Logit":
        # for models in statsmodels, the order of training and test datasets are reversed
        # and training and test datasets are defined in the model definition statement
        train_X, train_Y = def_var_names[1], def_var_names[0]
    else:
        if len(train_var_names) < 2:
            raise MutationException(f"Only {len(train_var_names)} train variable.")
        train_X, train_Y = train_var_names[0], train_var_names[1]
        if len(train_var_names) == 0:
            raise MutationException("No train variable found.")
    if len(test_var_names) == 0:
        raise MutationException("No train variable found.")
    test_X = test_var_names[0]
    assert task_type in ["classification", "regression"]

    return (
        m_type,
        task_type,
        def_stmt,
        train_stmt,
        test_stmt,
        model_pred_var,
        model_train_var,
        train_X.strip(),
        train_Y.strip(),
        test_X.strip(),
    )


# Creates mutated pipelines with the provided template models.
def model_mutation(in_path, out_dir, gen_mutation=True, gen_df_dumper=True):
    """Creates mutated pipelines with the provided template models.

    Parameters
    ----------
    in_path : PosixPath
        It contains path from pathlib.PosixPath.
    out_dir : PosixPath
        It contains path from pathlib.PosixPath.
    gen_mutation : bool
        True and otherwise False.
    gen_df_dumper : bool
        True and otherwise False.

    Returns
    -------
    m_type : str
        It returns a string.
    muted_types : list
        It returns a list.

    """
    if not os.path.exists(out_dir):
        os.makedirs(out_dir)

    (
        m_type,
        task_type,
        def_stmt,
        train_stmt,
        test_stmt,
        model_pred_var,
        model_train_var,
        train_X,
        train_Y,
        test_X,
    ) = pipeline_analysis(in_path)

    with open(in_path, "r", encoding="utf-8") as f:
        original_script = f.read()

    # first, dump current script to corresponding model type folder
    if not os.path.exists(out_dir / m_type):
        os.makedirs(out_dir / m_type)
    with open(out_dir / m_type / "script.py", "w", encoding="utf-8") as f:
        f.write(original_script)

    script_list = original_script.split("\n")

    muted_types = list()

    # then, start the main mutation script
    for key in templates:
        if not isinstance(key, tuple):
            continue
        mut_type, mut_task_type = key
        assert mut_type in order
        if mut_task_type == {"classification": "c", "regression": "r"}[task_type] and mut_type != m_type:
            value = templates[key]
            assert len(value) in [2, 3]
            if len(value) == 2:
                new_def_stmt, new_train_stmt = value
                ins_pred_stmt = None
            else:
                new_def_stmt, new_train_stmt, ins_pred_stmt = value

            if not os.path.exists(out_dir / mut_type):
                os.makedirs(out_dir / mut_type)

            new_script_list = script_list.copy()

            def_to_insert = (
                new_def_stmt.replace("#1", model_train_var)
                .replace("#2", model_pred_var)
                .replace("#XT", train_X)
                .replace("#YT", train_Y)
            )
            train_to_insert = (
                new_train_stmt.replace("#1", model_train_var)
                .replace("#2", model_pred_var)
                .replace("#XT", train_X)
                .replace("#YT", train_Y)
            )
            test_to_insert = None if ins_pred_stmt is None else ins_pred_stmt.replace("#XS", test_X)

            # index starts from 0
            t_def_stmt = list(def_stmt).copy()
            t_def_stmt[2] -= 1
            t_train_stmt = list(train_stmt).copy()
            t_train_stmt[2] -= 1
            t_test_stmt = list(test_stmt).copy()
            t_test_stmt[2] -= 1

            def_indent = (
                ""
                if len(script_list[t_def_stmt[2]].lstrip()) == len(script_list[t_def_stmt[2]])
                else script_list[t_def_stmt[2]][0]
                * (len(script_list[t_def_stmt[2]]) - len(script_list[t_def_stmt[2]].lstrip()))
            )
            train_indent = (
                ""
                if len(script_list[t_train_stmt[2]].lstrip()) == len(script_list[t_train_stmt[2]])
                else script_list[t_train_stmt[2]][0]
                * (len(script_list[t_train_stmt[2]]) - len(script_list[t_train_stmt[2]].lstrip()))
            )
            test_indent = (
                ""
                if len(script_list[t_test_stmt[2]].lstrip()) == len(script_list[t_test_stmt[2]])
                else script_list[t_test_stmt[2]][0]
                * (len(script_list[t_test_stmt[2]]) - len(script_list[t_test_stmt[2]].lstrip()))
            )
            def_to_insert = "\n".join([train_indent + item for item in def_to_insert.split("\n")])
            train_to_insert = "\n".join([def_indent + item for item in train_to_insert.split("\n")])
            test_to_insert = (
                script_list[t_test_stmt[2]]
                if test_to_insert is None
                else "\n".join(
                    [test_indent + item for item in test_to_insert.split("\n")] + [script_list[t_test_stmt[2]]]
                )
            )

            if t_def_stmt[2] == t_train_stmt[2]:
                new_script_list[t_def_stmt[2]] = "\n".join([def_to_insert, train_to_insert])
            else:
                new_script_list[t_def_stmt[2]] = def_to_insert
                new_script_list[t_train_stmt[2]] = train_to_insert
            new_script_list[t_test_stmt[2]] = test_to_insert

            # comment out column name printing lines and other unsupported specific statements
            for i, line in enumerate(new_script_list):
                if (
                    line.count("Y column Name") > 0
                    or line.count("Y column Type")
                    or line.count("summary()")
                    or line.count("cv_results_") > 0
                    or line.count("sm.fit_sample(x,y)") > 0
                ):
                    line_indent = "" if len(line.lstrip()) == len(line) else line[0] * (len(line) - len(line.lstrip()))
                    new_script_list[i] = line_indent + "pass"

            new_script = "\n".join(new_script_list)

            if gen_mutation:
                with open(out_dir / mut_type / "script.py", "w", encoding="utf-8") as f:
                    f.write(new_script)

            muted_types.append(mut_type)

    if gen_df_dumper:
        t_train_stmt = list(train_stmt).copy()
        t_train_stmt[2] -= 1
        train_indent = (
            ""
            if len(script_list[t_train_stmt[2]].lstrip()) == len(script_list[t_train_stmt[2]])
            else script_list[t_train_stmt[2]][0]
            * (len(script_list[t_train_stmt[2]]) - len(script_list[t_train_stmt[2]].lstrip()))
        )
        new_script_list = script_list.copy()
        dataset_pkl_folder = out_dir / "df_dump"
        dataset_pkl_path = out_dir / "df_dump" / "df.pkl"
        to_inject = f"""
{train_indent}import pandas as pd
{train_indent}import scipy.sparse as sparse
{train_indent}import numpy as np
{train_indent}_____train = {train_X}
# {train_indent}if isinstance(_____train, sparse.spmatrix): {train_X} = pd.DataFrame(_____train.toarray())
{train_indent}if isinstance(_____train, np.ndarray): _____train = pd.DataFrame(_____train)
{train_indent}if isinstance(_____train, sparse.spmatrix): _____train = pd.DataFrame.sparse.from_spmatrix(_____train)
{train_indent}if isinstance(_____train, pd.Series): _____train = _____train.to_frame()
{train_indent}_____cols = ['feature_' + str(col) for col in _____train.columns]
{train_indent}_____train.columns = _____cols--
{train_indent}_____target = {train_Y}
{train_indent}if isinstance(_____target, pd.Series): _____target = _____target.to_frame()
{train_indent}if isinstance(_____target, np.ndarray): _____target = pd.DataFrame(_____target)
{train_indent}if isinstance(_____target, list): _____target = pd.DataFrame(_____target)
{train_indent}_____cols = ['target_' + str(col) for col in _____target.columns]
{train_indent}_____target.columns = _____cols
{train_indent}_____merge = _____train.merge(_____target, left_index=True, right_index=True)
# {train_indent}if len(_____merge) > 1000000: _____merge = _____merge.iloc[numpy.random.choice(_____merge.shape[0], 1000000, replace=False)]
# {train_indent}for col in _____merge: _____merge[col] = _____merge[col].sparse.to_dense() if isinstance(_____merge[col].array, pd.SparseArray) else _____merge[col]
{train_indent}import os
{train_indent}if not os.path.exists(f'{dataset_pkl_folder}'): os.makedirs(f'{dataset_pkl_folder}')
{train_indent}_____merge.to_pickle(f"{dataset_pkl_path}")
{train_indent}exit()
"""
        new_script_list.insert(t_train_stmt[2], to_inject)
        new_script = "\n".join(new_script_list)
        if not os.path.exists(out_dir / "df_dump"):
            os.makedirs(out_dir / "df_dump")
        with open(out_dir / "df_dump" / "script.py", "w", encoding="utf-8") as f:
            f.write(new_script)

    return m_type, muted_types



# Copyright 2023-2024 The SapientML Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from typing import Optional

import libcst as cst
from libcst.metadata import ParentNodeProvider, PositionProvider


class SpecialVariableRetrieval(cst.CSTVisitor):
    METADATA_DEPENDENCIES = (
        PositionProvider,
        ParentNodeProvider,
    )

    def __init__(self):
        self.as_names = {}

    def visit_Assign(self, node) -> Optional[bool]:
        assigned_target = node.targets[0]
        target = assigned_target.target
        check = hasattr(target, "value")
        if check and target.value not in self.as_names:
            value = node.value
            if isinstance(value, cst.Call):
                if hasattr(value.func, "attr"):
                    api = str(value.func.attr.value)
                else:
                    api = str(value.func.value)
                if api == "read_csv":
                    self.as_names["dataframe"] = target.value


class ReferenceRetrieval(cst.CSTVisitor):
    METADATA_DEPENDENCIES = (
        PositionProvider,
        ParentNodeProvider,
    )

    def __init__(self):
        self.as_names = {}

    def visit_Assign(self, node) -> Optional[bool]:
        assigned_target = node.targets[0]
        target = assigned_target.target
        check = hasattr(target, "value")
        if check and target.value not in self.as_names:
            value = node.value
            if isinstance(value, cst.Call):
                pos = self.get_metadata(PositionProvider, node).start
                self.as_names[target.value] = pos.line


class ObjectRetrieval(cst.CSTVisitor):
    METADATA_DEPENDENCIES = (
        PositionProvider,
        ParentNodeProvider,
    )

    def __init__(self, function_name):
        self.object = ""
        self.function_name = function_name

    def visit_Call(self, node) -> Optional[bool]:
        if hasattr(node.func, "attr"):
            if node.func.attr.value == self.function_name:
                object_name = node.func.value.value
                self.object = object_name


class StringRetrieval(cst.CSTVisitor):
    METADATA_DEPENDENCIES = (
        PositionProvider,
        ParentNodeProvider,
    )

    def __init__(self):
        self.strings = set()

    def visit_SimpleString(self, node) -> Optional[bool]:
        string_value = node.value
        string_without_quote = string_value.replace("'", "")
        string_without_quote = string_without_quote.replace('"', "")
        self.strings.add(string_without_quote)


class PossibleColumnNamesApiMap(cst.CSTVisitor):
    METADATA_DEPENDENCIES = (
        PositionProvider,
        ParentNodeProvider,
    )

    def __init__(self):
        self.column_name_api = {}

    def visit_SimpleString(self, node) -> Optional[bool]:
        string_value = node.value
        string_without_quote = string_value.replace("'", "")
        string_without_quote = string_without_quote.replace('"', "")
        api = "n/a"
        parent = self.get_metadata(ParentNodeProvider, node)

        while True:
            if (
                isinstance(parent, cst.Call)
                or isinstance(parent, cst.SimpleStatementLine)
                or isinstance(parent, cst.Module)
            ):
                break
            parent = self.get_metadata(ParentNodeProvider, parent)
        if isinstance(parent, cst.Call):
            func = parent.func
            if isinstance(func, cst.Name):
                api = func.value
            elif isinstance(func, cst.Attribute):
                api = func.attr.value
        if string_without_quote in self.column_name_api:
            api_list = self.column_name_api[string_without_quote]
            api_list.append(api)
        else:
            self.column_name_api[string_without_quote] = [api]


class ApiRetrieval(cst.CSTVisitor):
    METADATA_DEPENDENCIES = (
        PositionProvider,
        ParentNodeProvider,
    )

    def __init__(self):
        self.object = ""
        self.function_names = {}

    def visit_Call(self, node) -> Optional[bool]:
        if hasattr(node.func, "attr"):
            api = str(node.func.attr.value)
        else:
            api = str(node.func.value)

        count = 0
        if api in self.function_names:
            count = self.function_names[api]
        self.function_names[api] = count + 1


class ImportMapRetrieval(cst.CSTVisitor):
    METADATA_DEPENDENCIES = (
        PositionProvider,
        ParentNodeProvider,
    )

    def __init__(self):
        self.as_names = {}
        self.uses = {}

    def visit_AsName_name(self, node) -> None:
        pos = self.get_metadata(PositionProvider, node).start
        self.as_names[node.name.value] = pos.line

    def visit_ImportAlias(self, node) -> Optional[bool]:
        pos = self.get_metadata(PositionProvider, node).start
        self.as_names[node.name.value] = pos.line

    def visit_Name(self, node) -> Optional[bool]:
        if node.value in self.as_names:
            pos = self.get_metadata(PositionProvider, node).start
            if node.value not in self.uses:
                use_set = set()
            else:
                use_set = self.uses[node.value]
            use_set.add(pos.line)
            self.uses[node.value] = use_set


def get_import_map(code):
    source_tree = cst.parse_module(code)
    retriever = ImportMapRetrieval()
    wrapper = cst.metadata.MetadataWrapper(source_tree)
    wrapper.visit(retriever)
    line_2_import_map = {}
    for key1 in retriever.uses:
        uses = retriever.uses[key1]
        for use in uses:
            if use not in line_2_import_map:
                import_set = set()
            else:
                import_set = line_2_import_map[use]
            import_set.add(retriever.as_names[key1])
            line_2_import_map[use] = import_set

    return line_2_import_map


def get_definition_line_map(code):
    source_tree = cst.parse_module(code)
    retriever = ReferenceRetrieval()
    wrapper = cst.metadata.MetadataWrapper(source_tree)
    wrapper.visit(retriever)
    return retriever.as_names


def get_column_api_map(code):
    source_tree = cst.parse_module(code)
    retriever = PossibleColumnNamesApiMap()
    wrapper = cst.metadata.MetadataWrapper(source_tree)
    wrapper.visit(retriever)
    return retriever.column_name_api


def get_object_name(code, function_name):
    source_tree = cst.parse_module(code)
    retriever = ObjectRetrieval(function_name)
    wrapper = cst.metadata.MetadataWrapper(source_tree)
    wrapper.visit(retriever)
    return retriever.object


def get_api_count(code):
    source_tree = cst.parse_module(code)
    retriever = ApiRetrieval()
    wrapper = cst.metadata.MetadataWrapper(source_tree)
    wrapper.visit(retriever)
    function_names = retriever.function_names
    for api in function_names:
        if function_names[api] > 1:
            return "P"
    return "S"


def get_strings(code):
    source_tree = cst.parse_module(code)
    retriever = StringRetrieval()
    wrapper = cst.metadata.MetadataWrapper(source_tree)
    wrapper.visit(retriever)
    return retriever.strings


def get_special_variable(code):
    source_tree = cst.parse_module(code)
    retriever = SpecialVariableRetrieval()
    wrapper = cst.metadata.MetadataWrapper(source_tree)
    wrapper.visit(retriever)
    return retriever.as_names



# Copyright 2023-2024 The SapientML Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import datetime


class Code_Template:
    """Code Template class."""

    def __init__(self):
        self.str_reverse = {"NOW": str(datetime.datetime.now())}

    def update(self, lines):
        """update method.

        Parameters
        ----------
        lines : list[str]
            A line in block code from jupyter content template.

        Returns
        -------
        out : list[str]
            Updated line in block code from jupyter content template.

        """
        out = []
        for line in lines:
            for key in self.str_reverse:
                line = line.replace(key, self.str_reverse[key])
            out.append(line)
        return out



# Copyright 2023-2024 The SapientML Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import json
import os

from sapientml.util.logging import setup_logger
from sapientml_core import internal_path
from sapientml_core.training import project_corpus
from sapientml_core.util import file_util

logger = setup_logger()


def compare_basic(project, static_info):
    """This function create summary for each pipeline from dataset_snapshot.

    Parameters
    ----------
    project : ProjectInfo
        This contain information of all the pipelines.

    static_info : dict
        The parameter contains {'drop_api': ['Name'], 'rename_api': [], 'target': 'Legendary'}

    Returns
    -------
    dict
       {
            'used_cols': ['Generation', 'Legendary'],
            'unmapped_cols': ['Type 2', 'Attack'],
            'new_cols': ['Fire', 'Water', 'Bug'],
            'target': 'Legendary',
            'deleted': ['Name'],
            'status': 'IN_PROGRESS',

        }

    """
    # load Initial Dataframe and Final Dataframe
    summary = {}
    project_root = internal_path.clean_dir
    pipeline_relative_path = project.pipeline_path[len(str(project_root)) + 1 :]
    feature_processing_history_file_path = (
        internal_path.training_cache / "dataset-snapshots" / pipeline_relative_path.replace(".py", "_col_dumper.json")
    )

    if not os.path.exists(feature_processing_history_file_path):
        return summary

    feature_processing_history = file_util.load_json(feature_processing_history_file_path)

    column_snapshots = []
    for _, column_info in feature_processing_history[0].items():
        if column_info[2] == "<class 'pandas.core.frame.DataFrame'>":
            if len(column_info[0]) > 1:
                no_int_present = True
                for column_name in column_info[0]:
                    if isinstance(column_name, int):
                        no_int_present = False
                        break
                if no_int_present:
                    column_snapshots.append(column_info[0])

    if len(column_snapshots) == 0:
        return summary

    # compare shapes
    summary["pipeline"] = project.file_name

    initial_features = set(column_snapshots[0])

    final_features = set(column_snapshots[-1])
    target = static_info.get("target", "n/a")
    if isinstance(target, list):
        final_features |= set(target)
    else:
        final_features.add(target)

    used_columns = final_features.intersection(initial_features)
    unmapped_cols = initial_features.difference(used_columns)
    new_cols = final_features.difference(used_columns)

    # column expansion detection
    column_expansion = set()
    remove_from_new_cols = set()
    for unmapped in unmapped_cols:
        for new_col in new_cols:
            if new_col.startswith(unmapped):
                column_expansion.add(unmapped)
                remove_from_new_cols.add(new_col)
                used_columns.add(unmapped)

    unmapped_cols = unmapped_cols.difference(column_expansion)
    new_cols = new_cols.difference(remove_from_new_cols)

    # renaming handling
    possibly_renamed = set()
    renamed_cols = static_info["rename_api"]
    for unmapped_col in unmapped_cols:
        if unmapped_col in renamed_cols:
            index = renamed_cols.index(unmapped_col)
            renamed = renamed_cols[index + 1]
            if renamed in new_cols:
                possibly_renamed.add(unmapped_col)
                new_cols.remove(renamed)
                used_columns.add(unmapped_col)

    unmapped_cols = unmapped_cols.difference(possibly_renamed)

    # target handling

    if "0" in new_cols and static_info["target"] in unmapped_cols:
        used_columns.add(target)
        unmapped_cols.remove(target)
        new_cols.remove("0")

    # possible deletion
    possibly_deleted = set()
    deleted_cols = static_info["drop_api"]
    for unmapped_col in unmapped_cols:
        if unmapped_col in deleted_cols:
            possibly_deleted.add(unmapped_col)

    unmapped_cols = unmapped_cols.difference(possibly_deleted)

    # delete index
    if "index" in new_cols:
        new_cols.remove("index")

    summary["used_cols"] = list(used_columns)
    summary["unmapped_cols"] = list(unmapped_cols)
    summary["new_cols"] = list(new_cols)
    summary["target"] = target
    summary["deleted"] = list(possibly_deleted)

    if len(new_cols) == 0 or len(unmapped_cols) == 0:
        summary["status"] = "FINALIZED"
    else:
        summary["status"] = "IN_PROGRESS"
        logger.info(file_util.read_file(project.pipeline_path))

    return summary


def main(test_mode=False):
    """This main function fetch all the pipeline details from json files in
    dataset_snapshots folder and summarize it and store in feature_analysis_summary.json
    Summary consist of following information:
    pipeline name
    used_cols
    unmapped_cols
    new_cols
    target
    deleted
    status

    Parameters
    ----------
    test_mode : bool
        The parameter is using for test mode.

    Raises
    ------
    Exception
            The ``Raises`` section is a list of all exceptions
            that are relevant to the interface.

    """
    # get selected pipelines
    corpus = project_corpus.ProjectCorpus()
    projects = corpus.project_list

    total_number_target_pipelines = len(projects)
    finalized = 0
    error = 0

    current_summary = {}

    try:
        with open(internal_path.training_cache / "static_info.json", "r", encoding="utf-8") as f:
            static_info = json.load(f)
    except Exception:
        raise

    feature_removed = 0

    for i in range(0, total_number_target_pipelines):
        if test_mode and i > 5:
            break
        pipeline = projects[i].pipeline_path
        file_name = projects[i].file_name
        if file_name not in static_info:
            continue

        logger.info(f"COMPARING: {i + 1} out of:{total_number_target_pipelines}, PIPELINE:{pipeline}")
        if file_name in current_summary and current_summary[file_name] is not None:
            status = current_summary[file_name].get("status", "n/a")
            if status == "FINALIZED":
                finalized += 1
                continue
        try:
            summary = compare_basic(projects[i], static_info[file_name])
            status = summary.get("status", "n/a")
            if status == "FINALIZED":
                finalized += 1

            deleted = summary.get("deleted", [])
            unmapped = summary.get("unmapped_cols", [])
            if len(deleted) + len(unmapped) > 0:
                feature_removed += 1
        except Exception:
            summary = {"pipeline": file_name, "status": "ERROR"}
            error += 1
            raise

        current_summary[file_name] = summary

    with open(
        internal_path.training_cache / "feature_analysis_summary.json",
        "w",
        encoding="utf-8",
    ) as f:
        json.dump(current_summary, f, indent=4)

    logger.info(f"FINALIZED: {finalized},Total Pipeline:{len(current_summary)},Feature Removed:{feature_removed}")
    logger.error(f"ERROR:, {error}")


if __name__ == "__main__":
    import argparse

    parser = argparse.ArgumentParser()
    parser.add_argument("--tag", type=str, help="Tag for output files and dirs.")
    args = parser.parse_args()
    if args.tag:
        internal_path.training_cache = internal_path.training_cache / args.tag
    test_mode = False
    main(test_mode)
